from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, ClassVar, Dict, List, Optional, Set, Union

from great_expectations.core.batch import Batch, BatchRequestBase  # noqa: TCH001
from great_expectations.core.domain import Domain  # noqa: TCH001
from great_expectations.core.expectation_configuration import (
    ExpectationConfiguration,  # noqa: TCH001
)
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.rule_based_profiler.builder import Builder
from great_expectations.rule_based_profiler.config import (
    ParameterBuilderConfig,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.parameter_builder import (
    ParameterBuilder,
    init_rule_parameter_builders,
)
from great_expectations.rule_based_profiler.parameter_container import (
    ParameterContainer,  # noqa: TCH001
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class ExpectationConfigurationBuilder(ABC, Builder):
    exclude_field_names: ClassVar[Set[str]] = Builder.exclude_field_names | {
        "validation_parameter_builders",
    }

    def __init__(
        self,
        expectation_type: str,
        validation_parameter_builder_configs: Optional[
            List[ParameterBuilderConfig]
        ] = None,
        data_context: Optional[AbstractDataContext] = None,
        **kwargs,
    ) -> None:
        """
        The ExpectationConfigurationBuilder will build ExpectationConfiguration objects for a Domain from the Rule.

        Args:
            expectation_type: the "expectation_type" argument of "ExpectationConfiguration" object to be emitted.
            validation_parameter_builder_configs: ParameterBuilder configurations, having whose outputs available (as
            fully-qualified parameter names) is pre-requisite for present ExpectationConfigurationBuilder instance.
            These "ParameterBuilder" configurations help build kwargs needed for this "ExpectationConfigurationBuilder"
            data_context: AbstractDataContext associated with this ExpectationConfigurationBuilder
            kwargs: additional arguments
        """

        super().__init__(data_context=data_context)

        self._expectation_type = expectation_type

        self._validation_parameter_builders = init_rule_parameter_builders(
            parameter_builder_configs=validation_parameter_builder_configs,
            data_context=self._data_context,
        )

        """
        Since ExpectationConfigurationBuilderConfigSchema allows arbitrary fields (as ExpectationConfiguration kwargs)
        to be provided, they must be all converted to public property accessors and/or public fields in order for all
        provisions by Builder, SerializableDictDot, and DictDot to operate properly in compliance with their interfaces.
        """
        for k, v in kwargs.items():
            setattr(self, k, v)
            logger.debug(
                f'Setting unknown kwarg ({k}, {v}) provided to constructor as argument in "{self.__class__.__name__}".'
            )

    def build_expectation_configuration(
        self,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
        batch_list: Optional[List[Batch]] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> ExpectationConfiguration:
        """
        Args:
            domain: Domain object that is context for execution of this ParameterBuilder object.
            variables: attribute name/value pairs
            parameters: Dictionary of ParameterContainer objects corresponding to all Domain objects in memory.
            batch_list: Explicit list of Batch objects to supply data at runtime.
            batch_request: Explicit batch_request used to supply data at runtime.
            runtime_configuration: Additional run-time settings (see "Validator.DEFAULT_RUNTIME_CONFIGURATION").

        Returns:
            ExpectationConfiguration object.
        """
        self.resolve_validation_dependencies(
            domain=domain,
            variables=variables,
            parameters=parameters,
            batch_list=batch_list,
            batch_request=batch_request,
            runtime_configuration=runtime_configuration,
        )

        return self._build_expectation_configuration(
            domain=domain, variables=variables, parameters=parameters
        )

    def resolve_validation_dependencies(
        self,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
        batch_list: Optional[List[Batch]] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> None:
        validation_parameter_builders: List[ParameterBuilder] = (
            self.validation_parameter_builders or []
        )

        validation_parameter_builder: ParameterBuilder
        for validation_parameter_builder in validation_parameter_builders:
            validation_parameter_builder.build_parameters(
                domain=domain,
                variables=variables,
                parameters=parameters,
                parameter_computation_impl=None,
                batch_list=batch_list,
                batch_request=batch_request,
                runtime_configuration=runtime_configuration,
            )

    @abstractmethod
    def _build_expectation_configuration(
        self,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> ExpectationConfiguration:
        pass

    @property
    def expectation_type(self) -> str:
        return self._expectation_type

    @property
    def validation_parameter_builders(self) -> Optional[List[ParameterBuilder]]:
        return self._validation_parameter_builders


def init_rule_expectation_configuration_builders(
    expectation_configuration_builder_configs: List[dict],
    data_context: Optional[AbstractDataContext] = None,
) -> List[ExpectationConfigurationBuilder]:
    expectation_configuration_builder_config: dict
    return [
        init_expectation_configuration_builder(
            expectation_configuration_builder_config=expectation_configuration_builder_config,
            data_context=data_context,
        )
        for expectation_configuration_builder_config in expectation_configuration_builder_configs
    ]


def init_expectation_configuration_builder(
    expectation_configuration_builder_config: Union[
        ExpectationConfigurationBuilder, dict
    ],
    data_context: Optional[AbstractDataContext] = None,
) -> ExpectationConfigurationBuilder:
    if not isinstance(expectation_configuration_builder_config, dict):
        expectation_configuration_builder_config = (
            expectation_configuration_builder_config.to_dict()
        )

    expectation_configuration_builder: ExpectationConfigurationBuilder = instantiate_class_from_config(
        config=expectation_configuration_builder_config,
        runtime_environment={"data_context": data_context},
        config_defaults={
            "class_name": "DefaultExpectationConfigurationBuilder",
            "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
        },
    )
    return expectation_configuration_builder



"""
This is a template for creating custom MulticolumnMapExpectations.
For detailed instructions on how to use it, please see:
    https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_multicolumn_map_expectations
"""

from typing import Optional

from great_expectations.core.expectation_configuration import ExpectationConfiguration
from great_expectations.exceptions import InvalidExpectationConfigurationError
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.expectation import MulticolumnMapExpectation
from great_expectations.expectations.metrics.map_metric_provider import (
    MulticolumnMapMetricProvider,
    multicolumn_condition_partial,
)


# This class defines a Metric to support your Expectation.
# For most MulticolumnMapExpectations, the main business logic for calculation will live in this class.
# <snippet name="tests/integration/docusaurus/expectations/examples/multicolumn_map_expectation_template.py MulticolumnValuesMatchSomeCriteria class_def">
class MulticolumnValuesMatchSomeCriteria(MulticolumnMapMetricProvider):
    # </snippet>
    # This is the id string that will be used to reference your metric.
    # <snippet name="tests/integration/docusaurus/expectations/examples/multicolumn_map_expectation_template.py metric_name">
    condition_metric_name = "METRIC NAME GOES HERE"
    # </snippet>
    # These point your metric at the provided keys to facilitate calculation
    condition_domain_keys = (
        "batch_id",
        "table",
        "column_list",
        "row_condition",
        "condition_parser",
        "ignore_row_if",
    )
    condition_value_keys = ()

    # This method implements the core logic for the PandasExecutionEngine
    # <snippet name="tests/integration/docusaurus/expectations/examples/multicolumn_map_expectation_template.py pandas">
    @multicolumn_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column_list, **kwargs):
        raise NotImplementedError

    # </snippet>

    # This method defines the business logic for evaluating your metric when using a SqlAlchemyExecutionEngine
    # @multicolumn_condition_partial(engine=SqlAlchemyExecutionEngine)
    # def _sqlalchemy(cls, column_list, **kwargs):
    #     raise NotImplementedError

    # This method defines the business logic for evaluating your metric when using a SparkDFExecutionEngine
    # @multicolumn_condition_partial(engine=SparkDFExecutionEngine)
    # def _spark(cls, column_list, **kwargs):
    #     raise NotImplementedError


# This class defines the Expectation itself
# <snippet name="tests/integration/docusaurus/expectations/examples/multicolumn_map_expectation_template.py ExpectMulticolumnValuesToMatchSomeCriteria class_def">
class ExpectMulticolumnValuesToMatchSomeCriteria(MulticolumnMapExpectation):
    # </snippet>
    # <snippet name="tests/integration/docusaurus/expectations/examples/multicolumn_map_expectation_template.py docstring">
    """TODO: Add a docstring here"""
    # </snippet>

    # These examples will be shown in the public gallery.
    # They will also be executed as unit tests for your Expectation.
    examples = []

    # This is the id string of the Metric used by this Expectation.
    # For most Expectations, it will be the same as the `condition_metric_name` defined in your Metric class above.
    # <snippet name="tests/integration/docusaurus/expectations/examples/multicolumn_map_expectation_template.py map_metric">
    map_metric = "METRIC NAME GOES HERE"
    # </snippet>

    # This is a list of parameter names that can affect whether the Expectation evaluates to True or False
    success_keys = (
        "column_list",
        "mostly",
    )

    # This dictionary contains default values for any parameters that should have default values
    default_kwarg_values = {}

    def validate_configuration(
        self, configuration: Optional[ExpectationConfiguration] = None
    ) -> None:
        """
        Validates that a configuration has been set, and sets a configuration if it has yet to be set. Ensures that
        necessary configuration arguments have been provided for the validation of the expectation.

        Args:
            configuration (OPTIONAL[ExpectationConfiguration]): \
                An optional Expectation Configuration entry that will be used to configure the expectation
        Returns:
            None. Raises InvalidExpectationConfigurationError if the config is not validated successfully
        """

        super().validate_configuration(configuration)
        configuration = configuration or self.configuration

        # # Check other things in configuration.kwargs and raise Exceptions if needed
        # try:
        #     assert (
        #         ...
        #     ), "message"
        #     assert (
        #         ...
        #     ), "message"
        # except AssertionError as e:
        #     raise InvalidExpectationConfigurationError(str(e))

    # This object contains metadata for display in the public Gallery
    # <snippet name="tests/integration/docusaurus/expectations/examples/multicolumn_map_expectation_template.py library_metadata">
    library_metadata = {
        "tags": [],  # Tags for this Expectation in the Gallery
        "contributors": [  # Github handles for all contributors to this Expectation.
            "@your_name_here",  # Don't forget to add your github handle here!
        ],
    }
    # </snippet>


if __name__ == "__main__":
    # <snippet name="tests/integration/docusaurus/expectations/examples/multicolumn_map_expectation_template.py diagnostics">
    ExpectMulticolumnValuesToMatchSomeCriteria().print_diagnostic_checklist()
# </snippet>



import os
from unittest import mock

import pytest
from ruamel.yaml import YAML

from great_expectations import DataContext
from great_expectations.core import ExpectationSuite
from great_expectations.core.batch import BatchRequest
from great_expectations.core.usage_statistics.events import UsageStatsEvents
from great_expectations.rule_based_profiler import RuleBasedProfilerResult
from great_expectations.rule_based_profiler.rule_based_profiler import RuleBasedProfiler
from great_expectations.validator.metric_configuration import MetricConfiguration

yaml = YAML()


def test_batches_are_accessible(
    multibatch_generic_csv_generator,
    multibatch_generic_csv_generator_context,
):
    """
    What does this test and why?
    Batches created in the multibatch_generic_csv_generator fixture should be available using the
    multibatch_generic_csv_generator_context
    This test most likely duplicates tests elsewhere, but it is more of a test of the configurable fixture.
    """

    context: DataContext = multibatch_generic_csv_generator_context
    data_relative_path = "../data"
    data_path = os.path.join(context.root_directory, data_relative_path)
    datasource_name = "generic_csv_generator"
    data_connector_name = "daily_data_connector"
    asset_name = "daily_data_asset"

    datasource = context.datasources[datasource_name]

    data_connector = datasource.data_connectors[data_connector_name]

    total_batches: int = 20
    file_list = multibatch_generic_csv_generator(
        data_path=data_path, num_event_batches=total_batches
    )

    assert (
        data_connector._get_data_reference_list_from_cache_by_data_asset_name(
            data_asset_name=asset_name
        )
        == file_list
    )

    batch_request_1 = BatchRequest(
        datasource_name="generic_csv_generator",
        data_connector_name="daily_data_connector",
        data_asset_name="daily_data_asset",
        data_connector_query={
            "index": -1,
        },
    )
    # Should give most recent batch
    validator_1 = context.get_validator(
        batch_request=batch_request_1,
        create_expectation_suite_with_name="my_expectation_suite_name_1",
    )
    metric_max = validator_1.get_metric(
        MetricConfiguration("column.max", metric_domain_kwargs={"column": "batch_num"})
    )
    assert metric_max == total_batches
    metric_value_set = validator_1.get_metric(
        MetricConfiguration(
            "column.distinct_values",
            metric_domain_kwargs={"column": "string_cardinality_3"},
        )
    )
    assert metric_value_set == {"category0", "category1", "category2"}

    batch_request_2 = BatchRequest(
        datasource_name="generic_csv_generator",
        data_connector_name="daily_data_connector",
        data_asset_name="daily_data_asset",
        data_connector_query={
            "index": -2,
        },
    )
    validator_2 = context.get_validator(
        batch_request=batch_request_2,
        create_expectation_suite_with_name="my_expectation_suite_name_2",
    )
    metric_max = validator_2.get_metric(
        MetricConfiguration("column.max", metric_domain_kwargs={"column": "batch_num"})
    )
    assert metric_max == total_batches - 1
    metric_value_set = validator_2.get_metric(
        MetricConfiguration(
            "column.distinct_values",
            metric_domain_kwargs={"column": "string_cardinality_3"},
        )
    )
    assert metric_value_set == {"category0", "category1", "category2"}

    for batch_num in range(1, total_batches + 1):
        batch_request = BatchRequest(
            datasource_name="generic_csv_generator",
            data_connector_name="daily_data_connector",
            data_asset_name="daily_data_asset",
            data_connector_query={
                "index": -batch_num,
            },
        )
        validator = context.get_validator(
            batch_request=batch_request,
            create_expectation_suite_with_name=f"my_expectation_suite_name__{batch_num}",
        )
        metric_max = validator.get_metric(
            MetricConfiguration(
                "column.max", metric_domain_kwargs={"column": "batch_num"}
            )
        )
        assert metric_max == (total_batches + 1) - batch_num
        metric_value_set = set(
            validator.get_metric(
                MetricConfiguration(
                    "column.distinct_values",
                    metric_domain_kwargs={"column": "string_cardinality_3"},
                )
            )
        )
        assert metric_value_set == {"category0", "category1", "category2"}


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
@pytest.mark.slow  # 2.04s
def test_profile_includes_citations(
    mock_emit,
    alice_columnar_table_single_batch_context,
    alice_columnar_table_single_batch,
):
    # Load data context
    data_context: DataContext = alice_columnar_table_single_batch_context

    # Load profiler configs & loop (run tests for each one)
    yaml_config: str = alice_columnar_table_single_batch["profiler_config"]

    # Instantiate Profiler
    profiler_config = yaml.load(yaml_config)
    # `class_name`/`module_name` are generally consumed through `instantiate_class_from_config`
    # so we need to manually remove those values if we wish to use the **kwargs instantiation pattern
    profiler_config.pop("class_name")

    profiler: RuleBasedProfiler = RuleBasedProfiler(
        **profiler_config,
        data_context=data_context,
    )

    # BatchRequest yielding exactly one batch
    alice_single_batch_data_batch_request: dict = {
        "datasource_name": "alice_columnar_table_single_batch_datasource",
        "data_connector_name": "alice_columnar_table_single_batch_data_connector",
        "data_asset_name": "alice_columnar_table_single_batch_data_asset",
    }

    result: RuleBasedProfilerResult = profiler.run(
        batch_request=alice_single_batch_data_batch_request
    )

    assert result.citation is not None and len(result.citation.keys()) > 0

    assert mock_emit.call_count == 43
    assert all(
        payload[0][0]["event"] == "data_context.get_batch_list"
        for payload in mock_emit.call_args_list[:-1]
    )

    # noinspection PyUnresolvedReferences
    actual_events: List[unittest.mock._Call] = mock_emit.call_args_list
    assert actual_events[-1][0][0]["event"] == UsageStatsEvents.RULE_BASED_PROFILER_RUN


@mock.patch(
    "great_expectations.core.usage_statistics.usage_statistics.UsageStatisticsHandler.emit"
)
@pytest.mark.slow  # 2.16s
def test_profile_get_expectation_suite(
    mock_emit,
    alice_columnar_table_single_batch_context,
    alice_columnar_table_single_batch,
):
    # Load data context
    data_context: DataContext = alice_columnar_table_single_batch_context

    # Load profiler configs & loop (run tests for each one)
    yaml_config: str = alice_columnar_table_single_batch["profiler_config"]

    # Instantiate Profiler
    profiler_config = yaml.load(yaml_config)
    # `class_name`/`module_name` are generally consumed through `instantiate_class_from_config`
    # so we need to manually remove those values if we wish to use the **kwargs instantiation pattern
    profiler_config.pop("class_name")

    profiler: RuleBasedProfiler = RuleBasedProfiler(
        **profiler_config,
        data_context=data_context,
    )

    # BatchRequest yielding exactly one batch
    alice_single_batch_data_batch_request: dict = {
        "datasource_name": "alice_columnar_table_single_batch_datasource",
        "data_connector_name": "alice_columnar_table_single_batch_data_connector",
        "data_asset_name": "alice_columnar_table_single_batch_data_asset",
    }

    result: RuleBasedProfilerResult = profiler.run(
        batch_request=alice_single_batch_data_batch_request
    )

    expectation_suite_name: str = "my_suite"

    suite: ExpectationSuite = result.get_expectation_suite(
        expectation_suite_name=expectation_suite_name
    )

    assert suite is not None and len(suite.expectations) > 0

    assert mock_emit.call_count == 44

    # noinspection PyUnresolvedReferences
    actual_events: List[unittest.mock._Call] = mock_emit.call_args_list
    assert (
        actual_events[-1][0][0]["event"]
        == UsageStatsEvents.RULE_BASED_PROFILER_RESULT_GET_EXPECTATION_SUITE
    )



from typing import Dict, List

import pytest

from great_expectations.core import ExpectationConfiguration
from great_expectations.core.domain import Domain
from great_expectations.rule_based_profiler.parameter_container import ParameterNode
from great_expectations.rule_based_profiler.rule import RuleOutput


@pytest.mark.unit
def test_rule_output_get_get_expectation_configurations(
    rule_output_for_rule_state_with_domains_and_parameters,
):
    expected_expectation_configurations: List[ExpectationConfiguration] = [
        ExpectationConfiguration(
            **{
                "expectation_type": "expect_my_validation",
                "kwargs": {
                    "column": "Age",
                },
                "meta": {},
            }
        ),
        ExpectationConfiguration(
            **{
                "expectation_type": "expect_my_validation",
                "kwargs": {
                    "column": "Date",
                },
                "meta": {},
            }
        ),
    ]

    rule_output: RuleOutput = rule_output_for_rule_state_with_domains_and_parameters
    expectation_configurations: List[
        ExpectationConfiguration
    ] = rule_output.get_expectation_configurations()

    assert expectation_configurations == expected_expectation_configurations


@pytest.mark.unit
def test_rule_output_get_fully_qualified_parameter_names_by_domain(
    rule_output_for_rule_state_with_domains_and_parameters,
):
    expected_fully_qualified_parameter_names_by_domain: Dict[Domain, List[str]] = {
        Domain(
            domain_type="column",
            domain_kwargs={
                "column": "Age",
            },
            details={},
            rule_name="my_rule",
        ): [
            "$mean",
        ],
        Domain(
            domain_type="column",
            domain_kwargs={
                "column": "Date",
            },
            details={},
            rule_name="my_rule",
        ): [
            "$parameter.weekly_taxi_fairs.mean_values",
            "$parameter.tolerances.mostly",
            "$parameter.tolerances.financial.usd",
            "$parameter.monthly_taxi_fairs.mean_values",
            "$parameter.date_strings.yyyy_mm_dd_hh_mm_ss_tz_date_format",
            "$parameter.date_strings.yyyy_mm_dd_date_format",
            "$parameter.date_strings.tolerances.max_num_conversion_attempts",
            "$parameter.date_strings.tolerances.max_abs_error_time_milliseconds",
            "$parameter.date_strings.mm_yyyy_dd_hh_mm_ss_tz_date_format",
            "$parameter.date_strings.mm_yyyy_dd_date_format",
            "$parameter.daily_taxi_fairs.mean_values",
            "$mean",
        ],
    }

    rule_output: RuleOutput = rule_output_for_rule_state_with_domains_and_parameters
    fully_qualified_parameter_names_by_domain: Dict[
        Domain, List[str]
    ] = rule_output.get_fully_qualified_parameter_names_by_domain()

    assert (
        fully_qualified_parameter_names_by_domain
        == expected_fully_qualified_parameter_names_by_domain
    )


# noinspection PyPep8Naming
@pytest.mark.unit
def test_rule_output_get_fully_qualified_parameter_names_for_domain_id(
    rule_output_for_rule_state_with_domains_and_parameters,
    column_Age_domain,
):
    expected_fully_qualified_parameter_names: List[str] = [
        "$mean",
    ]

    rule_output: RuleOutput = rule_output_for_rule_state_with_domains_and_parameters
    fully_qualified_parameter_names: List[
        str
    ] = rule_output.get_fully_qualified_parameter_names_for_domain_id(
        domain_id=column_Age_domain.id
    )

    assert fully_qualified_parameter_names == expected_fully_qualified_parameter_names


@pytest.mark.unit
def test_rule_output_get_parameter_values_for_fully_qualified_parameter_names_by_domain(
    rule_output_for_rule_state_with_domains_and_parameters,
):
    expected_parameter_values_for_fully_qualified_parameter_names_by_domain: Dict[
        Domain, Dict[str, ParameterNode]
    ] = {
        Domain(
            rule_name="my_rule", domain_type="column", domain_kwargs={"column": "Age"}
        ): {"$mean": 5.0},
        Domain(
            domain_type="column",
            domain_kwargs={"column": "Date"},
            rule_name="my_rule",
        ): {
            "$parameter.weekly_taxi_fairs.mean_values": {
                "value": [
                    {
                        "sunday": 71.43,
                        "monday": 74.35,
                        "tuesday": 42.3,
                        "wednesday": 42.3,
                        "thursday": 82.2,
                        "friday": 78.78,
                        "saturday": 91.39,
                    },
                    {
                        "sunday": 81.43,
                        "monday": 84.35,
                        "tuesday": 52.3,
                        "wednesday": 43.3,
                        "thursday": 22.2,
                        "friday": 98.78,
                        "saturday": 81.39,
                    },
                    {
                        "sunday": 61.43,
                        "monday": 34.35,
                        "tuesday": 82.3,
                        "wednesday": 72.3,
                        "thursday": 22.2,
                        "friday": 38.78,
                        "saturday": 51.39,
                    },
                    {
                        "sunday": 51.43,
                        "monday": 64.35,
                        "tuesday": 72.3,
                        "wednesday": 82.3,
                        "thursday": 22.2,
                        "friday": 98.78,
                        "saturday": 31.39,
                    },
                    {
                        "sunday": 72.43,
                        "monday": 77.35,
                        "tuesday": 46.3,
                        "wednesday": 47.3,
                        "thursday": 88.2,
                        "friday": 79.78,
                        "saturday": 93.39,
                    },
                    {
                        "sunday": 72.43,
                        "monday": 73.35,
                        "tuesday": 41.3,
                        "wednesday": 49.3,
                        "thursday": 80.2,
                        "friday": 78.78,
                        "saturday": 93.39,
                    },
                    {
                        "sunday": 74.43,
                        "monday": 78.35,
                        "tuesday": 49.3,
                        "wednesday": 43.3,
                        "thursday": 88.2,
                        "friday": 72.78,
                        "saturday": 97.39,
                    },
                    {
                        "sunday": 73.43,
                        "monday": 72.35,
                        "tuesday": 40.3,
                        "wednesday": 40.3,
                        "thursday": 89.2,
                        "friday": 77.78,
                        "saturday": 90.39,
                    },
                    {
                        "sunday": 72.43,
                        "monday": 73.35,
                        "tuesday": 45.3,
                        "wednesday": 44.3,
                        "thursday": 89.2,
                        "friday": 77.78,
                        "saturday": 96.39,
                    },
                    {
                        "sunday": 75.43,
                        "monday": 74.25,
                        "tuesday": 42.33,
                        "wednesday": 42.23,
                        "thursday": 82.21,
                        "friday": 78.76,
                        "saturday": 91.37,
                    },
                    {
                        "sunday": 71.43,
                        "monday": 74.37,
                        "tuesday": 42.3,
                        "wednesday": 42.32,
                        "thursday": 82.23,
                        "friday": 78.77,
                        "saturday": 91.49,
                    },
                    {
                        "sunday": 71.63,
                        "monday": 74.37,
                        "tuesday": 42.2,
                        "wednesday": 42.1,
                        "thursday": 82.29,
                        "friday": 78.79,
                        "saturday": 91.39,
                    },
                    {
                        "sunday": 71.42,
                        "monday": 74.33,
                        "tuesday": 42.33,
                        "wednesday": 42.34,
                        "thursday": 82.25,
                        "friday": 78.77,
                        "saturday": 91.69,
                    },
                    {
                        "sunday": 71.44,
                        "monday": 72.35,
                        "tuesday": 42.33,
                        "wednesday": 42.31,
                        "thursday": 82.29,
                        "friday": 78.68,
                        "saturday": 91.49,
                    },
                    {
                        "sunday": 71.44,
                        "monday": 74.32,
                        "tuesday": 42.32,
                        "wednesday": 42.32,
                        "thursday": 82.29,
                        "friday": 78.77,
                        "saturday": 91.49,
                    },
                    {
                        "sunday": 71.44,
                        "monday": 74.33,
                        "tuesday": 42.21,
                        "wednesday": 42.31,
                        "thursday": 82.27,
                        "friday": 78.74,
                        "saturday": 91.49,
                    },
                    {
                        "sunday": 71.33,
                        "monday": 74.25,
                        "tuesday": 42.31,
                        "wednesday": 42.03,
                        "thursday": 82.02,
                        "friday": 78.08,
                        "saturday": 91.38,
                    },
                    {
                        "sunday": 71.41,
                        "monday": 74.31,
                        "tuesday": 42.39,
                        "wednesday": 42.93,
                        "thursday": 82.92,
                        "friday": 78.75,
                        "saturday": 91.49,
                    },
                    {
                        "sunday": 72.43,
                        "monday": 73.35,
                        "tuesday": 42.3,
                        "wednesday": 32.3,
                        "thursday": 52.2,
                        "friday": 88.78,
                        "saturday": 81.39,
                    },
                    {
                        "sunday": 71.43,
                        "monday": 74.35,
                        "tuesday": 32.3,
                        "wednesday": 92.3,
                        "thursday": 72.2,
                        "friday": 74.78,
                        "saturday": 51.39,
                    },
                    {
                        "sunday": 72.43,
                        "monday": 64.35,
                        "tuesday": 52.3,
                        "wednesday": 42.39,
                        "thursday": 82.28,
                        "friday": 78.77,
                        "saturday": 91.36,
                    },
                    {
                        "sunday": 81.43,
                        "monday": 94.35,
                        "tuesday": 62.3,
                        "wednesday": 52.3,
                        "thursday": 92.2,
                        "friday": 88.78,
                        "saturday": 51.39,
                    },
                    {
                        "sunday": 21.43,
                        "monday": 34.35,
                        "tuesday": 42.34,
                        "wednesday": 62.3,
                        "thursday": 52.2,
                        "friday": 98.78,
                        "saturday": 81.39,
                    },
                    {
                        "sunday": 71.33,
                        "monday": 74.25,
                        "tuesday": 42.13,
                        "wednesday": 42.93,
                        "thursday": 82.82,
                        "friday": 78.78,
                        "saturday": 91.39,
                    },
                    {
                        "sunday": 72.43,
                        "monday": 73.35,
                        "tuesday": 44.3,
                        "wednesday": 45.3,
                        "thursday": 86.2,
                        "friday": 77.78,
                        "saturday": 98.39,
                    },
                    {
                        "sunday": 79.43,
                        "monday": 78.35,
                        "tuesday": 47.3,
                        "wednesday": 46.3,
                        "thursday": 85.2,
                        "friday": 74.78,
                        "saturday": 93.39,
                    },
                    {
                        "sunday": 71.42,
                        "monday": 74.31,
                        "tuesday": 42.0,
                        "wednesday": 42.1,
                        "thursday": 82.23,
                        "friday": 65.78,
                        "saturday": 91.26,
                    },
                    {
                        "sunday": 91.43,
                        "monday": 84.35,
                        "tuesday": 42.37,
                        "wednesday": 42.36,
                        "thursday": 82.25,
                        "friday": 78.74,
                        "saturday": 91.32,
                    },
                    {
                        "sunday": 71.33,
                        "monday": 74.45,
                        "tuesday": 42.35,
                        "wednesday": 42.36,
                        "thursday": 82.27,
                        "friday": 26.78,
                        "saturday": 71.39,
                    },
                    {
                        "sunday": 71.53,
                        "monday": 73.35,
                        "tuesday": 43.32,
                        "wednesday": 42.23,
                        "thursday": 82.32,
                        "friday": 78.18,
                        "saturday": 91.49,
                    },
                    {
                        "sunday": 71.53,
                        "monday": 74.25,
                        "tuesday": 52.3,
                        "wednesday": 52.3,
                        "thursday": 81.23,
                        "friday": 78.78,
                        "saturday": 78.39,
                    },
                ],
                "details": {"confidence": "high"},
            },
            "$parameter.tolerances.mostly": 0.91,
            "$parameter.tolerances.financial.usd": 1.0,
            "$parameter.monthly_taxi_fairs.mean_values": {
                "value": [
                    2.3,
                    9.8,
                    42.3,
                    8.1,
                    38.5,
                    53.7,
                    71.43,
                    16.34,
                    49.43,
                    74.35,
                    51.98,
                    46.42,
                    20.01,
                    69.44,
                    65.32,
                    8.83,
                    55.79,
                    82.2,
                    36.93,
                    83.78,
                    31.13,
                    76.93,
                    67.67,
                    25.12,
                    58.04,
                    79.78,
                    90.91,
                    15.26,
                    61.65,
                    78.78,
                    12.99,
                ],
                "details": {"confidence": "low"},
            },
            "$parameter.date_strings.yyyy_mm_dd_hh_mm_ss_tz_date_format": {
                "value": "%Y-%m-%d %H:%M:%S %Z",
                "details": {"confidence": 0.78},
            },
            "$parameter.date_strings.yyyy_mm_dd_date_format": {
                "value": "%Y-%m-%d",
                "details": {"confidence": 0.78},
            },
            "$parameter.date_strings.tolerances.max_num_conversion_attempts": 5,
            "$parameter.date_strings.tolerances.max_abs_error_time_milliseconds": 100,
            "$parameter.date_strings.mm_yyyy_dd_hh_mm_ss_tz_date_format": {
                "value": "%m-%Y-%d %H:%M:%S %Z",
                "details": {"confidence": 0.78},
            },
            "$parameter.date_strings.mm_yyyy_dd_date_format": {
                "value": "%m-%Y-%d",
                "details": {"confidence": 0.78},
            },
            "$parameter.daily_taxi_fairs.mean_values": {
                "value": {
                    "sunday": 71.43,
                    "monday": 74.35,
                    "tuesday": 42.3,
                    "wednesday": 42.3,
                    "thursday": 82.2,
                    "friday": 78.78,
                    "saturday": 91.39,
                },
                "details": {"confidence": "medium"},
            },
            "$mean": 0.65,
        },
    }

    rule_output: RuleOutput = rule_output_for_rule_state_with_domains_and_parameters
    parameter_values_for_fully_qualified_parameter_names_by_domain: Dict[
        Domain, Dict[str, ParameterNode]
    ] = rule_output.get_parameter_values_for_fully_qualified_parameter_names_by_domain()

    assert (
        parameter_values_for_fully_qualified_parameter_names_by_domain
        == expected_parameter_values_for_fully_qualified_parameter_names_by_domain
    )


# noinspection PyPep8Naming
@pytest.mark.unit
def test_rule_output_get_parameter_values_for_fully_qualified_parameter_names_for_domain_id(
    rule_output_for_rule_state_with_domains_and_parameters,
    column_Age_domain,
):
    expected_parameter_values_for_fully_qualified_parameter_names: ParameterNode = (
        ParameterNode(
            {
                "$mean": 5.0,
            }
        )
    )

    rule_output: RuleOutput = rule_output_for_rule_state_with_domains_and_parameters
    parameter_values_for_fully_qualified_parameter_names: Dict[
        str, ParameterNode
    ] = rule_output.get_parameter_values_for_fully_qualified_parameter_names_for_domain_id(
        domain_id=column_Age_domain.id
    )

    assert (
        parameter_values_for_fully_qualified_parameter_names
        == expected_parameter_values_for_fully_qualified_parameter_names
    )



import pandas as pd
import pytest

import great_expectations.exceptions.exceptions
from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.data_context import DataContext
from great_expectations.expectations.core.expect_column_values_to_be_in_set import (
    ExpectColumnValuesToBeInSet,
)


# <snippet name="tests/expectations/core/test_expect_column_values_to_be_in_set.py ExpectColumnValuesToBeTwoLetterCountryCode_class_def">
class ExpectColumnValuesToBeTwoLetterCountryCode(ExpectColumnValuesToBeInSet):
    default_kwarg_values = {
        "value_set": ["FR", "DE", "CH", "ES", "IT", "BE", "NL", "PL"],
    }


# </snippet>
def test_expect_column_values_to_be_in_set_fail(
    data_context_with_datasource_pandas_engine,
):
    context: DataContext = data_context_with_datasource_pandas_engine

    df = pd.DataFrame(
        {
            "a": [
                "2021-01-01",
                "2021-01-31",
                "2021-02-28",
                "2021-03-20",
                "2021-02-21",
                "2021-05-01",
                "2021-06-18",
            ]
        }
    )

    batch_request = RuntimeBatchRequest(
        datasource_name="my_datasource",
        data_connector_name="default_runtime_data_connector_name",
        data_asset_name="my_data_asset",
        runtime_parameters={"batch_data": df},
        batch_identifiers={"default_identifier_name": "my_identifier"},
    )
    validator = context.get_validator(
        batch_request=batch_request,
        create_expectation_suite_with_name="test",
    )

    result = validator.expect_column_values_to_be_in_set(
        column="a", value_set=["2021-06-18"]
    )
    assert result.success is False


def test_expect_column_values_in_set_pass(
    data_context_with_datasource_pandas_engine,
):
    context: DataContext = data_context_with_datasource_pandas_engine

    df = pd.DataFrame(
        {
            "a": [
                "2021-01-01",
                "2021-01-31",
                "2021-02-28",
                "2021-03-20",
                "2021-02-21",
                "2021-05-01",
                "2021-06-18",
            ]
        }
    )

    batch_request = RuntimeBatchRequest(
        datasource_name="my_datasource",
        data_connector_name="default_runtime_data_connector_name",
        data_asset_name="my_data_asset",
        runtime_parameters={"batch_data": df},
        batch_identifiers={"default_identifier_name": "my_identifier"},
    )
    validator = context.get_validator(
        batch_request=batch_request,
        create_expectation_suite_with_name="test",
    )

    result = validator.expect_column_values_to_be_in_set(
        column="a",
        value_set=[
            "2021-01-01",
            "2021-01-31",
            "2021-02-28",
            "2021-03-20",
            "2021-02-21",
            "2021-05-01",
            "2021-06-18",
        ],
    )
    assert result.success is True


def test_expect_column_values_country_fail(
    data_context_with_datasource_pandas_engine,
):
    context: DataContext = data_context_with_datasource_pandas_engine

    df = pd.DataFrame(
        {
            "a": [
                "2021-01-01",
                "2021-01-31",
                "2021-02-28",
                "2021-03-20",
                "2021-02-21",
                "2021-05-01",
                "2021-06-18",
            ]
        }
    )

    batch_request = RuntimeBatchRequest(
        datasource_name="my_datasource",
        data_connector_name="default_runtime_data_connector_name",
        data_asset_name="my_data_asset",
        runtime_parameters={"batch_data": df},
        batch_identifiers={"default_identifier_name": "my_identifier"},
    )
    validator = context.get_validator(
        batch_request=batch_request,
        create_expectation_suite_with_name="test",
    )

    result = validator.expect_column_values_to_be_two_letter_country_code(column="a")

    assert result.success is False


def test_expect_column_values_country_pass(
    data_context_with_datasource_pandas_engine,
):
    context: DataContext = data_context_with_datasource_pandas_engine

    df = pd.DataFrame({"a": ["FR", "DE", "CH", "ES", "IT", "BE", "NL", "PL"]})

    batch_request = RuntimeBatchRequest(
        datasource_name="my_datasource",
        data_connector_name="default_runtime_data_connector_name",
        data_asset_name="my_data_asset",
        runtime_parameters={"batch_data": df},
        batch_identifiers={"default_identifier_name": "my_identifier"},
    )
    validator = context.get_validator(
        batch_request=batch_request,
        create_expectation_suite_with_name="test",
    )

    result = validator.expect_column_values_to_be_two_letter_country_code(column="a")

    assert result.success is True


def test_expect_column_values_to_be_in_set_no_set(
    data_context_with_datasource_pandas_engine,
):
    context: DataContext = data_context_with_datasource_pandas_engine

    df = pd.DataFrame(
        {
            "a": [
                "2021-01-01",
                "2021-01-31",
                "2021-02-28",
                "2021-03-20",
                "2021-02-21",
                "2021-05-01",
                "2021-06-18",
            ]
        }
    )

    batch_request = RuntimeBatchRequest(
        datasource_name="my_datasource",
        data_connector_name="default_runtime_data_connector_name",
        data_asset_name="my_data_asset",
        runtime_parameters={"batch_data": df},
        batch_identifiers={"default_identifier_name": "my_identifier"},
    )
    validator = context.get_validator(
        batch_request=batch_request,
        create_expectation_suite_with_name="test",
    )
    with pytest.raises(
        great_expectations.exceptions.exceptions.InvalidExpectationConfigurationError
    ):
        result = validator.expect_column_values_to_be_in_set(column="a")



import os

from ruamel import yaml

import great_expectations as gx

context = gx.get_context()

yaml = yaml.YAML(typ="safe")

# parse great_expectations.yml for comparison
great_expectations_yaml_file_path = os.path.join(
    context.root_directory, "great_expectations.yml"
)
with open(great_expectations_yaml_file_path) as f:
    great_expectations_yaml = yaml.load(f)

actual_datasource = great_expectations_yaml["datasources"]

# expected Datasource
expected_existing_datasource_yaml = r"""
  my_datasource:
    module_name: great_expectations.datasource
    class_name: Datasource
    execution_engine:
      module_name: great_expectations.execution_engine
      class_name: SparkDFExecutionEngine
    data_connectors:
      default_inferred_data_connector_name:
        class_name: InferredAssetFilesystemDataConnector
        base_directory: ../../../data/
        module_name: great_expectations.datasource.data_connector
        batch_spec_passthrough:
          reader_options:
            header: True
        default_regex:
          group_names:
            - data_asset_name
          pattern: (.*)
      default_runtime_data_connector_name:
        class_name: RuntimeDataConnector
        batch_identifiers:
          - default_identifier_name
        module_name: great_expectations.datasource.data_connector
"""

assert actual_datasource == yaml.load(expected_existing_datasource_yaml)

# Please note this override is only to provide good UX for docs and tests.
updated_configuration = yaml.load(expected_existing_datasource_yaml)
updated_configuration["my_datasource"]["data_connectors"][
    "default_inferred_data_connector_name"
]["base_directory"] = "../data/"
context.add_datasource(name="my_datasource", **updated_configuration["my_datasource"])

# check that checkpoint contains the right configuration
# parse great_expectations.yml for comparison
checkpoint_yaml_file_path = os.path.join(
    context.root_directory, "checkpoints/test_v3_checkpoint.yml"
)
with open(checkpoint_yaml_file_path) as f:
    actual_checkpoint_yaml = yaml.load(f)

expected_checkpoint_yaml = """
name: test_v3_checkpoint
config_version: 1.0 # Note this is the version of the Checkpoint configuration, and not the great_expectations.yml configuration
template_name:
module_name: great_expectations.checkpoint
class_name: Checkpoint
run_name_template: '%Y%m%d-%H%M%S-my-run-name-template'
expectation_suite_name:
batch_request:
action_list:
  - name: store_validation_result
    action:
      class_name: StoreValidationResultAction
  - name: store_evaluation_params
    action:
      class_name: StoreEvaluationParametersAction
  - name: update_data_docs
    action:
      class_name: UpdateDataDocsAction
      site_names: []
evaluation_parameters: {}
runtime_configuration: {}
validations:
  - batch_request:
      datasource_name: my_datasource
      data_connector_name: default_inferred_data_connector_name
      data_asset_name: Titanic.csv
      data_connector_query:
        index: -1
    expectation_suite_name: Titanic.profiled
profilers: []
"""

assert actual_checkpoint_yaml == yaml.load(expected_checkpoint_yaml)

# run checkpoint
context.add_or_update_checkpoint(**actual_checkpoint_yaml)
results = context.run_checkpoint(checkpoint_name="test_v3_checkpoint")

assert results["success"] is True



import click

from great_expectations.cli import toolkit
from great_expectations.cli.pretty_printing import cli_message, cli_message_dict
from great_expectations.core.usage_statistics.events import UsageStatsEvents
from great_expectations.core.usage_statistics.util import send_usage_message


@click.group()
@click.pass_context
def store(ctx: click.Context) -> None:
    """Store operations"""
    ctx.obj.data_context = ctx.obj.get_data_context_from_config_file()

    cli_event_noun: str = "store"
    (
        begin_event_name,
        end_event_name,
    ) = UsageStatsEvents.get_cli_begin_and_end_event_names(
        noun=cli_event_noun,
        verb=ctx.invoked_subcommand,  # type: ignore[arg-type]
    )
    send_usage_message(
        data_context=ctx.obj.data_context,
        event=begin_event_name,
        success=True,
    )
    ctx.obj.usage_event_end = end_event_name


@store.command(name="list")
@click.pass_context
def store_list(ctx: click.Context):
    """List active Stores."""
    context = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end
    try:
        stores = context.list_active_stores()
        cli_message(f"{len(stores)} active Stores found:")
        for store in stores:
            cli_message("")
            cli_message_dict(store)

        send_usage_message(
            data_context=context,
            event=usage_event_end,
            success=True,
        )
    except Exception as e:
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message=f"<red>{e}</red>",
        )
        return



"""Example Script: How to configure a Spark/Pandas/Sql Datasource (universal configuration elements)

This example script is intended for use in documentation on how to configure Datasources.  It contains the top level
configuration items that are identical for Spark, Pandas, and Sql.

Assert statements are included to ensure that if the behaviour shown in this script breaks it will not pass
tests and will be updated.  These statements can be ignored by users.

Comments with the tags `<snippet>` and `</snippet>` are used to ensure that if this script is updated
the snippets that are specified for use in documentation are maintained.  These comments can be ignored by users.

--documentation--
    https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/datasource_configuration/how_to_configure_a_pandas_datasource
    https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/datasource_configuration/how_to_configure_a_spark_datasource
    https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/datasource_configuration/how_to_configure_a_sql_datasource
"""

# The following imports are used as part of verifying that all example snippets are consistent.
# Users may disregard them.

from datasource_configuration_test_utilities import is_subset
from full_datasource_configurations import (
    get_partial_config_universal_datasource_config_elements,
)

import great_expectations as gx

# The following methods correspond to the section headings in the how-to guide linked in the module docstring.


def section_1_import_necessary_modules_and_initialize_your_data_context() -> gx.DataContext:
    """Provides and tests the snippets for section 1 of the Spark, Pandas, and SQL Datasource configuration guides.

    Returns:
        a Great Expectations DataContext object
    """
    # <snippet name="import necessary modules and initialize your data context">
    from ruamel import yaml

    import great_expectations as gx

    data_context: gx.DataContext = gx.get_context()
    # </snippet>

    assert isinstance(data_context, gx.DataContext)
    return data_context


def section_2_create_new_datasource_configuration():
    """Provides and tests the snippets for section 2 of the Spark, Pandas, and SQL Datasource configuration guides."""
    # <snippet name="create empty datasource_config dictionary">
    datasource_config: dict = {}
    # </snippet>
    is_subset(
        datasource_config, get_partial_config_universal_datasource_config_elements()
    )


def section_3_name_your_datasource():
    """Provides and tests the snippets for section 3 of the Spark, Pandas, and SQL Datasource configuration guides."""
    # <snippet name="datasource_config up to name being populated as my_datasource_name">
    datasource_config: dict = {"name": "my_datasource_name"}
    # </snippet>

    name_snippet: dict = {
        # <snippet name="populate name as my_datasource_name">
        "name": "my_datasource_name"
        # </snippet>
    }
    assert name_snippet == datasource_config
    is_subset(
        datasource_config, get_partial_config_universal_datasource_config_elements()
    )


def section_4_specify_the_datasource_class_and_module():
    """Provides and tests the snippets for section 4 of the Spark, Pandas, and SQL Datasource configuration guides."""
    # <snippet name="datasource_config containing top level elements universal to spark pandas and sql Datasources">
    datasource_config: dict = {
        "name": "my_datasource_name",  # Preferably name it something relevant
        "class_name": "Datasource",
        "module_name": "great_expectations.datasource",
    }
    # </snippet>
    assert (
        datasource_config == get_partial_config_universal_datasource_config_elements()
    )
    is_subset(
        datasource_config, get_partial_config_universal_datasource_config_elements()
    )


section_1_import_necessary_modules_and_initialize_your_data_context()
section_2_create_new_datasource_configuration()
section_3_name_your_datasource()
section_4_specify_the_datasource_class_and_module()



import os
import tempfile

import great_expectations as gx
from great_expectations.data_context.types.base import (
    DataContextConfig,
    DatasourceConfig,
    FilesystemStoreBackendDefaults,
)
from great_expectations.util import get_context

"""
A simple test to verify that `context.build_data_docs()` works as expected.

As indicated in issue #3772, calling `context.build_data_docs()` raised an unexpected exception
when Great Expectations was installed in a non-filesystem location (i.e. it failed when
GX was installed inside a zip file -which is a location allowed by PEP 273-).

Therefore, this test is intended to be run after installing GX inside a zip file and
then setting the appropriate PYTHONPATH env variable. If desired, this test can also be
run after installing GX in a normal filesystem location (i.e. a directory).

This test is OK if it finishes without raising an exception.

To make it easier to debug this test, it prints:
* The location of the GX library: to verify that we are testing the library that we want
* The version of the GX library: idem
* data_docs url: If everything works, this will be a url (e.g. starting with file://...)


Additional info: https://github.com/great-expectations/great_expectations/issues/3772 and
https://www.python.org/dev/peps/pep-0273/
"""

print(f"Great Expectations location: {gx.__file__}")
print(f"Great Expectations version: {gx.__version__}")

data_context_config = DataContextConfig(
    datasources={"example_datasource": DatasourceConfig(class_name="PandasDatasource")},
    store_backend_defaults=FilesystemStoreBackendDefaults(
        root_directory=tempfile.mkdtemp() + os.sep + "my_greatexp_workdir"
    ),
)
context = get_context(project_config=data_context_config)
print(f"Great Expectations data_docs url: {context.build_data_docs()}")



from __future__ import annotations

import copy
import datetime
import json
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union, cast
from uuid import UUID

import great_expectations.exceptions as gx_exceptions
from great_expectations.checkpoint.configurator import SimpleCheckpointConfigurator
from great_expectations.checkpoint.types.checkpoint_result import CheckpointResult
from great_expectations.checkpoint.util import (
    batch_request_in_validations_contains_batch_data,
    get_substituted_validation_dict,
    get_validations_with_batch_request_as_dict,
    substitute_runtime_config,
    substitute_template_config,
)
from great_expectations.core import RunIdentifier
from great_expectations.core._docs_decorators import (
    deprecated_argument,
    new_argument,
    public_api,
)
from great_expectations.core.async_executor import AsyncExecutor, AsyncResult
from great_expectations.core.batch import (
    BatchRequest,
    BatchRequestBase,
    RuntimeBatchRequest,
    batch_request_contains_batch_data,
    get_batch_request_as_dict,
)
from great_expectations.core.config_peer import ConfigOutputModes, ConfigPeer
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,  # noqa: TCH001
    ExpectationSuiteValidationResultMeta,  # noqa: TCH001
)
from great_expectations.core.usage_statistics.events import UsageStatsEvents
from great_expectations.core.usage_statistics.usage_statistics import (
    get_checkpoint_run_usage_statistics,
    usage_statistics_enabled_method,
)
from great_expectations.data_asset import DataAsset
from great_expectations.data_context.cloud_constants import (
    CLOUD_APP_DEFAULT_BASE_URL,
    GXCloudRESTResource,
)
from great_expectations.data_context.types.base import (
    CheckpointConfig,
    CheckpointValidationConfig,
)
from great_expectations.data_context.types.resource_identifiers import GXCloudIdentifier
from great_expectations.data_context.util import instantiate_class_from_config
from great_expectations.util import (
    deep_filter_properties_iterable,
    filter_properties_dict,
    load_class,
)
from great_expectations.validation_operators import ActionListValidationOperator
from great_expectations.validation_operators.types.validation_operator_result import (
    ValidationOperatorResult,  # noqa: TCH001
)

if TYPE_CHECKING:
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.validator.validator import Validator


logger = logging.getLogger(__name__)


def _get_validator_class() -> Type[Validator]:
    """Using this function helps work around circular import dependncies."""
    module_name = "great_expectations.validator.validator"
    class_name = "Validator"
    return load_class(class_name=class_name, module_name=module_name)


class BaseCheckpoint(ConfigPeer):
    """
    BaseCheckpoint class is initialized from CheckpointConfig typed object and contains all functionality
    in the form of interface methods (which can be overwritten by subclasses) and their reference implementation.
    """

    def __init__(
        self,
        checkpoint_config: CheckpointConfig,
        data_context: AbstractDataContext,
    ) -> None:
        from great_expectations.data_context.data_context.abstract_data_context import (
            AbstractDataContext,
        )

        if not isinstance(data_context, AbstractDataContext):
            raise TypeError("A Checkpoint requires a valid DataContext")

        self._usage_statistics_handler = data_context._usage_statistics_handler

        self._data_context = data_context

        self._checkpoint_config = checkpoint_config

    # TODO: Add eval param processing using new TBD parser syntax and updated EvaluationParameterParser and
    #  parse_evaluation_parameters function (e.g. datetime substitution or specifying relative datetimes like "most
    #  recent"). Currently, environment variable substitution is the only processing applied to evaluation parameters,
    #  while run_name_template also undergoes strftime datetime substitution
    @public_api
    @usage_statistics_enabled_method(
        event_name=UsageStatsEvents.CHECKPOINT_RUN,
        args_payload_fn=get_checkpoint_run_usage_statistics,
    )
    @new_argument(
        argument_name="expectation_suite_ge_cloud_id",
        version="0.13.33",
        message="Used in cloud deployments.",
    )
    def run(
        self,
        template_name: Optional[str] = None,
        run_name_template: Optional[str] = None,
        expectation_suite_name: Optional[str] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        action_list: Optional[List[dict]] = None,
        evaluation_parameters: Optional[dict] = None,
        runtime_configuration: Optional[dict] = None,
        validations: Optional[List[dict]] = None,
        profilers: Optional[List[dict]] = None,
        run_id: Optional[Union[str, RunIdentifier]] = None,
        run_name: Optional[str] = None,
        run_time: Optional[Union[str, datetime.datetime]] = None,
        result_format: Optional[Union[str, dict]] = None,
        expectation_suite_ge_cloud_id: Optional[str] = None,
    ) -> CheckpointResult:
        """Validate against current Checkpoint.

        Arguments allow for override of the current Checkpoint configuration.

        Args:
            template_name: The name of another checkpoint to use as a base template.
            run_name_template: A template to create run names, using environment
                variables and datetime-template syntax (e.g. "%Y-%M-staging-$MY_ENV_VAR").
            expectation_suite_name: Expectation suite associated with checkpoint.
            batch_request: Batch request describing the batch of data to validate.
            action_list: A list of actions to perform after each batch is validated.
            evaluation_parameters: Evaluation parameters to use in generating this checkpoint.
            runtime_configuration: Runtime configuration to pass into the validator's runtime configuration
                (e.g. `result_format`).
            validations: Validations to be executed as part of checkpoint.
            profilers: Profilers to use in generating this checkpoint.
            run_id: The run_id for the validation; if None, a default value will be used.
            run_name: The run_name for the validation; if None, a default value will be used.
            run_time: The date/time of the run.
            result_format: One of several supported formatting directives for expectation validation results
            expectation_suite_ge_cloud_id: Great Expectations Cloud id for the expectation suite

        Raises:
            InvalidCheckpointConfigError: If `run_id` is provided with `run_name` or `run_time`.
            InvalidCheckpointConfigError: If `result_format` is not an expected type.
            CheckpointError: If Checkpoint does not contain a `batch_request` or validations.

        Returns:
            CheckpointResult
        """
        if (run_id and run_name) or (run_id and run_time):
            gx_exceptions.InvalidCheckpointConfigError(
                "Please provide either a run_id or run_name and/or run_time"
            )

        # If no validations are provided, the combination of expectation_suite_name, batch_request,
        # and action_list are considered the "default" validation.
        using_default_validation = not self.validations and not validations

        run_time = run_time or datetime.datetime.now()
        runtime_configuration = runtime_configuration or {}
        result_format = result_format or runtime_configuration.get("result_format")

        _result_format_types = (type(None), str, dict)
        if not isinstance(result_format, _result_format_types):
            raise gx_exceptions.InvalidCheckpointConfigError(
                f"result_format should be of type - {' '.join(str(t) for t in _result_format_types)}"
            )

        batch_request = get_batch_request_as_dict(batch_request=batch_request)
        validations = get_validations_with_batch_request_as_dict(
            validations=validations
        )

        runtime_kwargs: dict = {
            "template_name": template_name,
            "run_name_template": run_name_template,
            "expectation_suite_name": expectation_suite_name,
            "batch_request": batch_request or {},
            "action_list": action_list or [],
            "evaluation_parameters": evaluation_parameters or {},
            "runtime_configuration": runtime_configuration or {},
            "validations": validations or [],
            "profilers": profilers or [],
            "expectation_suite_ge_cloud_id": expectation_suite_ge_cloud_id,
        }

        substituted_runtime_config: dict = self.get_substituted_config(
            runtime_kwargs=runtime_kwargs
        )

        run_name_template = substituted_runtime_config.get("run_name_template")

        batch_request = substituted_runtime_config.get("batch_request")
        validations = cast(list, substituted_runtime_config.get("validations") or [])

        if len(validations) == 0 and not batch_request:
            raise gx_exceptions.CheckpointError(
                f'Checkpoint "{self.name}" must contain either a batch_request or validations.'
            )

        if run_name is None and run_name_template is not None:
            run_name = run_time.strftime(run_name_template)

        run_id = run_id or RunIdentifier(run_name=run_name, run_time=run_time)

        # Ensure that validations dicts have the most specific id available
        # (default to Checkpoint's default_validation_id if no validations were passed in the signature)
        if using_default_validation:
            for validation in validations:
                validation["id"] = self.config.default_validation_id

        # Use AsyncExecutor to speed up I/O bound validations by running them in parallel with multithreading (if
        # concurrency is enabled in the data context configuration) -- please see the below arguments used to initialize
        # AsyncExecutor and the corresponding AsyncExecutor docstring for more details on when multiple threads are
        # used.
        with AsyncExecutor(
            self.data_context.concurrency, max_workers=len(validations)
        ) as async_executor:
            # noinspection PyUnresolvedReferences
            async_validation_operator_results: List[
                AsyncResult[ValidationOperatorResult]
            ] = []
            if len(validations) > 0:
                for idx, validation_dict in enumerate(validations):
                    self._run_validation(
                        substituted_runtime_config=substituted_runtime_config,
                        async_validation_operator_results=async_validation_operator_results,
                        async_executor=async_executor,
                        result_format=result_format,
                        run_id=run_id,
                        idx=idx,
                        validation_dict=validation_dict,
                    )
            else:
                self._run_validation(
                    substituted_runtime_config=substituted_runtime_config,
                    async_validation_operator_results=async_validation_operator_results,
                    async_executor=async_executor,
                    result_format=result_format,
                    run_id=run_id,
                )

            checkpoint_run_results: dict = {}
            async_validation_operator_result: AsyncResult
            for async_validation_operator_result in async_validation_operator_results:
                async_result = async_validation_operator_result.result()
                run_results = async_result.run_results

                run_result: dict
                validation_result: Optional[ExpectationSuiteValidationResult]
                meta: ExpectationSuiteValidationResultMeta
                for run_result in run_results.values():
                    validation_result = run_result.get("validation_result")
                    if validation_result:
                        meta = validation_result.meta
                        id = str(self.ge_cloud_id) if self.ge_cloud_id else None
                        meta["checkpoint_id"] = id

                checkpoint_run_results.update(run_results)

        # Generate a URL to the validation result details page in GX Cloud
        validation_result_url: str | None = None
        for key in checkpoint_run_results:
            if isinstance(key, GXCloudIdentifier) and key.cloud_id:
                validation_result_url = (
                    f"{CLOUD_APP_DEFAULT_BASE_URL}?validationResultId={key.cloud_id}"
                )
                break
        return CheckpointResult(
            validation_result_url=validation_result_url,
            run_id=run_id,
            run_results=checkpoint_run_results,
            checkpoint_config=self.config,
        )

    def get_substituted_config(
        self,
        runtime_kwargs: Optional[dict] = None,
    ) -> dict:
        if runtime_kwargs is None:
            runtime_kwargs = {}

        config_kwargs: dict = self.get_config(mode=ConfigOutputModes.JSON_DICT)

        template_name: Optional[str] = runtime_kwargs.get("template_name")
        if template_name:
            config_kwargs["template_name"] = template_name

        substituted_runtime_config: dict = self._get_substituted_template(
            source_config=config_kwargs
        )
        substituted_runtime_config = self._get_substituted_runtime_kwargs(
            source_config=substituted_runtime_config, runtime_kwargs=runtime_kwargs
        )

        return substituted_runtime_config

    def _get_substituted_template(
        self,
        source_config: dict,
    ) -> dict:
        substituted_config: dict

        template_name = source_config.get("template_name")
        if template_name:
            checkpoint: Checkpoint = self.data_context.get_checkpoint(
                name=template_name
            )
            template_config: dict = checkpoint.config.to_json_dict()

            if template_config["config_version"] != source_config["config_version"]:
                raise gx_exceptions.CheckpointError(
                    f"Invalid template '{template_name}' (ver. {template_config['config_version']}) for Checkpoint "
                    f"'{source_config}' (ver. {source_config['config_version']}. Checkpoints can only use templates with the same config_version."
                )

            substituted_template_config: dict = self._get_substituted_template(
                source_config=template_config
            )
            substituted_config = substitute_template_config(
                source_config=source_config, template_config=substituted_template_config
            )
        else:
            substituted_config = copy.deepcopy(source_config)

        if self._using_cloud_context:
            return substituted_config

        return self._substitute_config_variables(config=substituted_config)

    def _get_substituted_runtime_kwargs(
        self,
        source_config: dict,
        runtime_kwargs: Optional[dict] = None,
    ) -> dict:
        if runtime_kwargs is None:
            runtime_kwargs = {}

        substituted_config: dict = substitute_runtime_config(
            source_config=source_config, runtime_kwargs=runtime_kwargs
        )

        if self._using_cloud_context:
            return substituted_config

        return self._substitute_config_variables(config=substituted_config)

    def _substitute_config_variables(self, config: dict) -> dict:
        return self.data_context.config_provider.substitute_config(config)

    def _run_validation(
        self,
        substituted_runtime_config: dict,
        async_validation_operator_results: List[AsyncResult],
        async_executor: AsyncExecutor,
        result_format: Optional[dict],
        run_id: Optional[Union[str, RunIdentifier]],
        idx: Optional[int] = 0,
        validation_dict: Optional[dict] = None,
    ) -> None:
        if validation_dict is None:
            validation_dict = {}
            validation_dict["id"] = substituted_runtime_config.get(
                "default_validation_id"
            )

        try:
            substituted_validation_dict: dict = get_substituted_validation_dict(
                substituted_runtime_config=substituted_runtime_config,
                validation_dict=validation_dict,
            )
            batch_request: Union[
                BatchRequest, RuntimeBatchRequest
            ] = substituted_validation_dict.get("batch_request")
            expectation_suite_name: str = substituted_validation_dict.get(
                "expectation_suite_name"
            )
            expectation_suite_ge_cloud_id: str = substituted_validation_dict.get(
                "expectation_suite_ge_cloud_id"
            )
            include_rendered_content: Optional[bool] = substituted_validation_dict.get(
                "include_rendered_content"
            )
            if include_rendered_content is None:
                include_rendered_content = (
                    self._data_context._determine_if_expectation_validation_result_include_rendered_content()
                )

            validator: Validator = self.data_context.get_validator(
                batch_request=batch_request,
                expectation_suite_name=expectation_suite_name
                if not self._using_cloud_context
                else None,
                expectation_suite_ge_cloud_id=(
                    expectation_suite_ge_cloud_id if self._using_cloud_context else None
                ),
                include_rendered_content=include_rendered_content,
            )

            action_list: list = substituted_validation_dict.get("action_list")
            runtime_configuration_validation = substituted_validation_dict.get(
                "runtime_configuration", {}
            )
            catch_exceptions_validation = runtime_configuration_validation.get(
                "catch_exceptions"
            )
            result_format_validation = runtime_configuration_validation.get(
                "result_format"
            )
            result_format = result_format or result_format_validation

            if result_format is None:
                result_format = {"result_format": "SUMMARY"}

            action_list_validation_operator: ActionListValidationOperator = (
                ActionListValidationOperator(
                    data_context=self.data_context,
                    action_list=action_list,
                    result_format=result_format,
                    name=f"{self.name}-checkpoint-validation[{idx}]",
                )
            )
            checkpoint_identifier = None
            if self._using_cloud_context:
                checkpoint_identifier = GXCloudIdentifier(
                    resource_type=GXCloudRESTResource.CHECKPOINT,
                    cloud_id=str(self.ge_cloud_id),
                )

            operator_run_kwargs = {}

            if catch_exceptions_validation is not None:
                operator_run_kwargs["catch_exceptions"] = catch_exceptions_validation

            validation_id: Optional[str] = substituted_validation_dict.get("id")

            async_validation_operator_result = async_executor.submit(
                action_list_validation_operator.run,
                assets_to_validate=[validator],
                run_id=run_id,
                evaluation_parameters=substituted_validation_dict.get(
                    "evaluation_parameters"
                ),
                result_format=result_format,
                checkpoint_identifier=checkpoint_identifier,
                checkpoint_name=self.name,
                validation_id=validation_id,
                **operator_run_kwargs,
            )
            async_validation_operator_results.append(async_validation_operator_result)
        except (
            gx_exceptions.CheckpointError,
            gx_exceptions.ExecutionEngineError,
            gx_exceptions.MetricError,
        ) as e:
            raise gx_exceptions.CheckpointError(
                f"Exception occurred while running validation[{idx}] of Checkpoint '{self.name}': {e.message}."
            ) from e

    def self_check(self, pretty_print: bool = True) -> dict:
        """Method that is intended to provide visibility into parameters that Checkpoint was instantiated with.

        If used as part of the test_yaml_config() workflow, the user will be able to know if the Checkpoint is
        configured with all necessary parameters.

        When run with self_check()::

            yaml_config: str = # [a checkpoint yaml configuration]
            config: CommentedMap = yaml.load(yaml_config)
            checkpoint_config: CheckpointConfig = CheckpointConfig(**config)
            checkpoint: Checkpoint = Checkpoint(
                data_context=context,
                checkpoint_config.to_json_dict()
            )
            checkpoint.self_check()

        When run with test_yaml_config::

            checkpoint: Checkpoint = context.test_yaml_config(
                yaml_config=yaml_config,
                name="my_checkpoint"
                )

        Args:
            pretty_print (bool): If True, then additional messages if Checkpoint configuration is missing
                a "validations" or "action_list" attribute.

        Returns:
            Dictionary containing Checkpoint configuration converted into json dictionary.

        """
        report_object: dict = {"config": self.config.to_json_dict()}

        if pretty_print:
            print(f"\nCheckpoint class name: {self.__class__.__name__}")

        validations_present: bool = (
            self.validations
            and isinstance(self.validations, list)
            and len(self.validations) > 0
        )
        action_list: Optional[list] = self.action_list
        action_list_present: bool = (
            action_list is not None
            and isinstance(action_list, list)
            and len(action_list) > 0
        ) or (
            validations_present
            and all(
                [
                    (
                        validation.get("action_list")
                        and isinstance(validation["action_list"], list)
                        and len(validation["action_list"]) > 0
                    )
                    for validation in self.validations
                ]
            )
        )
        if pretty_print:
            if not validations_present:
                print(
                    """Your current Checkpoint configuration has an empty or missing "validations" attribute.  This
means you must either update your Checkpoint configuration or provide an appropriate validations
list programmatically (i.e., when your Checkpoint is run).
                    """
                )
            if not action_list_present:
                print(
                    """Your current Checkpoint configuration has an empty or missing "action_list" attribute.  This
means you must provide an appropriate validations list programmatically (i.e., when your Checkpoint
is run), with each validation having its own defined "action_list" attribute.
                    """
                )

        return report_object

    @property
    def config(self) -> CheckpointConfig:
        return self._checkpoint_config

    @property
    def name(self) -> Optional[str]:
        try:
            return self.config.name
        except AttributeError:
            return None

    @property
    def config_version(self) -> Optional[float]:
        try:
            return self.config.config_version
        except AttributeError:
            return None

    @property
    def action_list(self) -> List[Dict]:
        try:
            return self.config.action_list
        except AttributeError:
            return []

    @property
    def validations(self) -> List[CheckpointValidationConfig]:
        try:
            return self.config.validations
        except AttributeError:
            return []

    @property
    def ge_cloud_id(self) -> Optional[UUID]:
        try:
            return self.config.ge_cloud_id
        except AttributeError:
            return None

    @property
    def data_context(self) -> AbstractDataContext:
        return self._data_context

    @property
    def _using_cloud_context(self) -> bool:
        # Chetan - 20221216 - This is a temporary property to encapsulate any Cloud leakage
        # Upon refactoring this class to decouple Cloud-specific branches, this should be removed
        from great_expectations.data_context.data_context.cloud_data_context import (
            CloudDataContext,
        )

        return isinstance(self.data_context, CloudDataContext)

    def __repr__(self) -> str:
        return str(self.get_config())


@public_api
@deprecated_argument(argument_name="validation_operator_name", version="0.14.0")
@deprecated_argument(argument_name="batches", version="0.14.0")
@new_argument(
    argument_name="ge_cloud_id", version="0.13.33", message="Used in cloud deployments."
)
@new_argument(
    argument_name="expectation_suite_ge_cloud_id",
    version="0.13.33",
    message="Used in cloud deployments.",
)
class Checkpoint(BaseCheckpoint):
    """A checkpoint is the primary means for validating data in a production deployment of Great Expectations.

    Checkpoints provide a convenient abstraction for bundling the Validation of a Batch (or Batches) of data against
    an Expectation Suite (or several), as well as the Actions that should be taken after the validation.

    A Checkpoint uses a Validator to run one or more Expectation Suites against one or more Batches provided by one
    or more Batch Requests. Running a Checkpoint produces Validation Results and will result in optional Actions
    being performed if they are configured to do so.

    Args:
        name: User-selected cCheckpoint name (e.g. "staging_tables").
        data_context: Data context that is associated with the current checkpoint.
        config_version: Version number of the checkpoint configuration.
        template_name: The name of another checkpoint to use as a base template.
        run_name_template: A template to create run names, using environment
            variables and datetime-template syntax (e.g. "%Y-%M-staging-$MY_ENV_VAR").
        expectation_suite_name: Expectation suite associated with checkpoint.
        batch_request: Batch request describing the batch of data to validate.
        action_list: A list of actions to perform after each batch is validated.
        evaluation_parameters: Evaluation parameters to use in generating this checkpoint.
        runtime_configuration: Runtime configuration to pass into the validator's runtime configuration
            (e.g. `result_format`).
        validations: Validations to be executed as part of checkpoint.
        profilers: Profilers to use in generating this checkpoint.
        validation_operator_name: List of validation Operators configured by the Checkpoint.
        batches: List of Batches for validation by Checkpoint.
        ge_cloud_id: Great Expectations Cloud id for this Checkpoint.
        expectation_suite_ge_cloud_id: Great Expectations Cloud id associated with Expectation Suite.
        default_validation_id:  Default value used by Checkpoint if no Validations are configured.

    Raises:
        ValueError: If BatchRequest contains batch_data, since only primitive types are allowed in the constructor.
        ValueError: If Validations contains batch_data, since only primitive types are allowed in the constructor.
    """

    """
    --ge-feature-maturity-info--

        id: checkpoint
        title: Newstyle Class-based Checkpoints
        short_description: Run a configured checkpoint from a notebook.
        description: Run a configured checkpoint from a notebook.
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/guides/how_to_guides/validation/how_to_create_a_new_checkpoint.html
        maturity: Beta
        maturity_details:
            api_stability: Mostly stable (transitioning ValidationOperators to Checkpoints)
            implementation_completeness: Complete
            unit_test_coverage: Partial ("golden path"-focused tests; error checking tests need to be improved)
            integration_infrastructure_test_coverage: N/A
            documentation_completeness: Complete
            bug_risk: Medium

    --ge-feature-maturity-info--
    """

    def __init__(
        self,
        name: str,
        data_context: AbstractDataContext,
        config_version: Optional[Union[int, float]] = None,
        template_name: Optional[str] = None,
        run_name_template: Optional[str] = None,
        expectation_suite_name: Optional[str] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        action_list: Optional[List[dict]] = None,
        evaluation_parameters: Optional[dict] = None,
        runtime_configuration: Optional[dict] = None,
        validations: Optional[List[dict]] = None,
        profilers: Optional[List[dict]] = None,
        validation_operator_name: Optional[str] = None,
        batches: Optional[List[dict]] = None,
        ge_cloud_id: Optional[UUID] = None,
        expectation_suite_ge_cloud_id: Optional[UUID] = None,
        default_validation_id: Optional[str] = None,
    ) -> None:
        # Only primitive types are allowed as constructor arguments; data frames are supplied to "run()" as arguments.
        if batch_request_contains_batch_data(batch_request=batch_request):
            raise ValueError(
                """Error: batch_data found in batch_request -- only primitive types are allowed as Checkpoint \
constructor arguments.
"""
            )

        if batch_request_in_validations_contains_batch_data(validations=validations):
            raise ValueError(
                """Error: batch_data found in batch_request -- only primitive types are allowed as Checkpoint \
constructor arguments.
"""
            )

        checkpoint_config = CheckpointConfig(
            name=name,
            config_version=config_version,
            template_name=template_name,
            run_name_template=run_name_template,
            expectation_suite_name=expectation_suite_name,
            batch_request=batch_request,
            action_list=action_list,
            evaluation_parameters=evaluation_parameters,
            runtime_configuration=runtime_configuration,
            validations=validations,
            profilers=profilers,
            validation_operator_name=validation_operator_name,
            batches=batches,
            ge_cloud_id=ge_cloud_id,
            expectation_suite_ge_cloud_id=expectation_suite_ge_cloud_id,
            default_validation_id=default_validation_id,
        )
        super().__init__(
            checkpoint_config=checkpoint_config,
            data_context=data_context,
        )

    def run_with_runtime_args(
        self,
        template_name: Optional[str] = None,
        run_name_template: Optional[str] = None,
        expectation_suite_name: Optional[str] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        action_list: Optional[List[dict]] = None,
        evaluation_parameters: Optional[dict] = None,
        runtime_configuration: Optional[dict] = None,
        validations: Optional[List[dict]] = None,
        profilers: Optional[List[dict]] = None,
        run_id: Optional[Union[str, int, float]] = None,
        run_name: Optional[str] = None,
        run_time: Optional[datetime.datetime] = None,
        result_format: Optional[str] = None,
        expectation_suite_ge_cloud_id: Optional[str] = None,
        **kwargs,
    ) -> CheckpointResult:
        checkpoint_config_from_store: CheckpointConfig = cast(
            CheckpointConfig, self.get_config()
        )

        if (
            "runtime_configuration" in checkpoint_config_from_store
            and checkpoint_config_from_store.runtime_configuration
            and "result_format" in checkpoint_config_from_store.runtime_configuration
        ):
            result_format = (
                result_format
                or checkpoint_config_from_store.runtime_configuration.get(
                    "result_format"
                )
            )

        if result_format is None:
            result_format = {"result_format": "SUMMARY"}

        batch_request = get_batch_request_as_dict(batch_request=batch_request)
        validations = get_validations_with_batch_request_as_dict(
            validations=validations
        )

        checkpoint_config_from_call_args: dict = {
            "template_name": template_name,
            "run_name_template": run_name_template,
            "expectation_suite_name": expectation_suite_name,
            "batch_request": batch_request,
            "action_list": action_list,
            "evaluation_parameters": evaluation_parameters,
            "runtime_configuration": runtime_configuration,
            "validations": validations,
            "profilers": profilers,
            "run_id": run_id,
            "run_name": run_name,
            "run_time": run_time,
            "result_format": result_format,
            "expectation_suite_ge_cloud_id": expectation_suite_ge_cloud_id,
        }

        checkpoint_config: dict = {
            key: value
            for key, value in checkpoint_config_from_store.items()
            if key in checkpoint_config_from_call_args
        }
        checkpoint_config.update(checkpoint_config_from_call_args)

        checkpoint_run_arguments: dict = dict(**checkpoint_config, **kwargs)
        filter_properties_dict(
            properties=checkpoint_run_arguments,
            clean_falsy=True,
            inplace=True,
        )

        return self.run(**checkpoint_run_arguments)

    @staticmethod
    def construct_from_config_args(
        data_context: AbstractDataContext,
        checkpoint_store_name: str,
        name: str,
        config_version: Optional[Union[int, float]] = None,
        template_name: Optional[str] = None,
        module_name: Optional[str] = None,
        class_name: Optional[str] = None,
        run_name_template: Optional[str] = None,
        expectation_suite_name: Optional[str] = None,
        batch_request: Optional[dict] = None,
        action_list: Optional[List[dict]] = None,
        evaluation_parameters: Optional[dict] = None,
        runtime_configuration: Optional[dict] = None,
        validations: Optional[List[dict]] = None,
        profilers: Optional[List[dict]] = None,
        # Next two fields are for LegacyCheckpoint configuration
        validation_operator_name: Optional[str] = None,
        batches: Optional[List[dict]] = None,
        # the following four arguments are used by SimpleCheckpoint
        site_names: Optional[Union[str, List[str]]] = None,
        slack_webhook: Optional[str] = None,
        notify_on: Optional[str] = None,
        notify_with: Optional[Union[str, List[str]]] = None,
        ge_cloud_id: Optional[str] = None,
        expectation_suite_ge_cloud_id: Optional[str] = None,
        default_validation_id: Optional[str] = None,
    ) -> Checkpoint:
        checkpoint_config: Union[CheckpointConfig, dict]

        # These checks protect against typed objects (BatchRequest and/or RuntimeBatchRequest) encountered in arguments.
        batch_request = get_batch_request_as_dict(batch_request=batch_request)
        validations = get_validations_with_batch_request_as_dict(
            validations=validations
        )

        # DataFrames shouldn't be saved to CheckpointStore
        if batch_request_contains_batch_data(batch_request=batch_request):
            raise gx_exceptions.InvalidConfigError(
                f'batch_data found in batch_request cannot be saved to CheckpointStore "{checkpoint_store_name}"'
            )

        if batch_request_in_validations_contains_batch_data(validations=validations):
            raise gx_exceptions.InvalidConfigError(
                f'batch_data found in validations cannot be saved to CheckpointStore "{checkpoint_store_name}"'
            )

        checkpoint_config = {
            "name": name,
            "config_version": config_version,
            "template_name": template_name,
            "module_name": module_name,
            "class_name": class_name,
            "run_name_template": run_name_template,
            "expectation_suite_name": expectation_suite_name,
            "batch_request": batch_request,
            "action_list": action_list,
            "evaluation_parameters": evaluation_parameters,
            "runtime_configuration": runtime_configuration,
            "validations": validations,
            "profilers": profilers,
            # Next two fields are for LegacyCheckpoint configuration
            "validation_operator_name": validation_operator_name,
            "batches": batches,
            # the following four keys are used by SimpleCheckpoint
            "site_names": site_names,
            "slack_webhook": slack_webhook,
            "notify_on": notify_on,
            "notify_with": notify_with,
            "ge_cloud_id": ge_cloud_id,
            "expectation_suite_ge_cloud_id": expectation_suite_ge_cloud_id,
            "default_validation_id": default_validation_id,
        }

        checkpoint_config = deep_filter_properties_iterable(
            properties=checkpoint_config,
            clean_falsy=True,
        )

        new_checkpoint: Checkpoint = instantiate_class_from_config(
            config=checkpoint_config,
            runtime_environment={
                "data_context": data_context,
            },
            config_defaults={
                "module_name": "great_expectations.checkpoint",
            },
        )

        return new_checkpoint

    @staticmethod
    def instantiate_from_config_with_runtime_args(
        checkpoint_config: CheckpointConfig,
        data_context: AbstractDataContext,
        **runtime_kwargs,
    ) -> Checkpoint:
        config: dict = checkpoint_config.to_json_dict()

        key: str
        value: Any
        for key, value in runtime_kwargs.items():
            if value is not None:
                config[key] = value

        config = filter_properties_dict(properties=config, clean_falsy=True)

        checkpoint: Checkpoint = instantiate_class_from_config(
            config=config,
            runtime_environment={
                "data_context": data_context,
            },
            config_defaults={
                "module_name": "great_expectations.checkpoint",
            },
        )

        return checkpoint


class LegacyCheckpoint(Checkpoint):
    """
    --ge-feature-maturity-info--

        id: checkpoint_notebook
        title: LegacyCheckpoint - Notebook
        icon:
        short_description: Run a configured Checkpoint from a notebook.
        description: Run a configured Checkpoint from a notebook.
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/guides/how_to_guides/validation/how_to_run_a_checkpoint_in_python.html
        maturity: Experimental (to-be-deprecated in favor of Checkpoint)
        maturity_details:
            api_stability: to-be-deprecated in favor of Checkpoint
            implementation_completeness: Complete
            unit_test_coverage: Partial ("golden path"-focused tests; error checking tests need to be improved)
            integration_infrastructure_test_coverage: N/A
            documentation_completeness: Complete
            bug_risk: Low

        id: checkpoint_command_line
        title: LegacyCheckpoint - Command Line
        icon:
        short_description: Run a configured Checkpoint from a command line.
        description: Run a configured checkpoint from a command line in a Terminal shell.
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/guides/how_to_guides/validation/how_to_run_a_checkpoint_in_terminal.html
        maturity: Experimental (to-be-deprecated in favor of Checkpoint)
        maturity_details:
            api_stability: to-be-deprecated in favor of Checkpoint
            implementation_completeness: Complete
            unit_test_coverage: Complete
            integration_infrastructure_test_coverage: N/A
            documentation_completeness: Complete
            bug_risk: Low

        id: checkpoint_cron_job
        title: LegacyCheckpoint - Cron
        icon:
        short_description: Deploy a configured Checkpoint as a scheduled task with cron.
        description: Use the Unix crontab command to edit the cron file and add a line that will run Checkpoint as a scheduled task.
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/guides/how_to_guides/validation/how_to_deploy_a_scheduled_checkpoint_with_cron.html
        maturity: Experimental (to-be-deprecated in favor of Checkpoint)
        maturity_details:
            api_stability: to-be-deprecated in favor of Checkpoint
            implementation_completeness: Complete
            unit_test_coverage: Complete
            integration_infrastructure_test_coverage: N/A
            documentation_completeness: Complete
            bug_risk: Low

        id: checkpoint_airflow_dag
        title: LegacyCheckpoint - Airflow DAG
        icon:
        short_description: Run a configured Checkpoint in Apache Airflow
        description: Running a configured Checkpoint in Apache Airflow enables the triggering of data validation using an Expectation Suite directly within an Airflow DAG.
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/guides/how_to_guides/validation/how_to_run_a_checkpoint_in_airflow.html
        maturity: Beta (to-be-deprecated in favor of Checkpoint)
        maturity_details:
            api_stability: to-be-deprecated in favor of Checkpoint
            implementation_completeness: Partial (no operator, but probably don't need one)
            unit_test_coverage: N/A
            integration_infrastructure_test_coverage: Minimal
            documentation_completeness: Complete (pending how-to)
            bug_risk: Low

        id: checkpoint_kedro
        title: LegacyCheckpoint - Kedro
        icon:
        short_description:
        description:
        how_to_guide_url:
        maturity: Experimental (to-be-deprecated in favor of Checkpoint)
        maturity_details:
            api_stability: to-be-deprecated in favor of Checkpoint
            implementation_completeness: Unknown
            unit_test_coverage: Unknown
            integration_infrastructure_test_coverage: Unknown
            documentation_completeness:  Minimal (none)
            bug_risk: Unknown

        id: checkpoint_prefect
        title: LegacyCheckpoint - Prefect
        icon:
        short_description:
        description:
        how_to_guide_url:
        maturity: Experimental (to-be-deprecated in favor of Checkpoint)
        maturity_details:
            api_stability: to-be-deprecated in favor of Checkpoint
            implementation_completeness: Unknown
            unit_test_coverage: Unknown
            integration_infrastructure_test_coverage: Unknown
            documentation_completeness: Minimal (none)
            bug_risk: Unknown

        id: checkpoint_dbt
        title: LegacyCheckpoint - DBT
        icon:
        short_description:
        description:
        how_to_guide_url:
        maturity: Beta (to-be-deprecated in favor of Checkpoint)
        maturity_details:
            api_stability: to-be-deprecated in favor of Checkpoint
            implementation_completeness: Minimal
            unit_test_coverage: Minimal (none)
            integration_infrastructure_test_coverage: Minimal (none)
            documentation_completeness: Minimal (none)
            bug_risk: Low

    --ge-feature-maturity-info--
    """

    def __init__(
        self,
        name: str,
        data_context,
        validation_operator_name: Optional[str] = None,
        batches: Optional[List[dict]] = None,
    ) -> None:
        super().__init__(
            name=name,
            data_context=data_context,
            validation_operator_name=validation_operator_name,
            batches=batches,
        )

        self._validation_operator_name = validation_operator_name
        self._batches = batches

    @property
    def validation_operator_name(self) -> Optional[str]:
        return self._validation_operator_name

    @property
    def batches(self) -> Optional[List[dict]]:
        return self._batches

    def _run_default_validation_operator(
        self,
        assets_to_validate: List,
        run_id: Optional[Union[str, RunIdentifier]] = None,
        evaluation_parameters: Optional[dict] = None,
        run_name: Optional[str] = None,
        run_time: Optional[Union[str, datetime.datetime]] = None,
        result_format: Optional[Union[str, dict]] = None,
    ) -> ValidationOperatorResult:
        result_format = result_format or {"result_format": "SUMMARY"}

        if not assets_to_validate:
            raise gx_exceptions.DataContextError(
                "No batches of data were passed in. These are required"
            )

        for batch in assets_to_validate:
            if not isinstance(batch, (tuple, DataAsset, _get_validator_class())):
                raise gx_exceptions.DataContextError(
                    "Batches are required to be of type DataAsset or Validator"
                )

        if run_id is None and run_name is None:
            run_name = datetime.datetime.now(datetime.timezone.utc).strftime(
                "%Y%m%dT%H%M%S.%fZ"
            )
            logger.info(f"Setting run_name to: {run_name}")

        default_validation_operator = ActionListValidationOperator(
            data_context=self.data_context,
            action_list=[
                {
                    "name": "store_validation_result",
                    "action": {"class_name": "StoreValidationResultAction"},
                },
                {
                    "name": "store_evaluation_params",
                    "action": {"class_name": "StoreEvaluationParametersAction"},
                },
                {
                    "name": "update_data_docs",
                    "action": {"class_name": "UpdateDataDocsAction", "site_names": []},
                },
            ],
            result_format=result_format,
            name="default-action-list-validation-operator",
        )

        if evaluation_parameters is None:
            return default_validation_operator.run(
                assets_to_validate=assets_to_validate,
                run_id=run_id,
                run_name=run_name,
                run_time=run_time,
                result_format=result_format,
            )
        else:
            return default_validation_operator.run(
                assets_to_validate=assets_to_validate,
                run_id=run_id,
                evaluation_parameters=evaluation_parameters,
                run_name=run_name,
                run_time=run_time,
                result_format=result_format,
            )

    def run(
        self,
        run_id=None,
        evaluation_parameters=None,
        run_name=None,
        run_time=None,
        result_format=None,
        **kwargs,
    ):
        """Legacy Checkpoint that has been deprecated since 0.14.0.

        Args:
            run_id: The run_id for the validation; if None, a default value will be used.
            evaluation_parameters: Evaluation parameters to use in generating this checkpoint.
            run_name: The run_name for the validation; if None, a default value will be used.
            run_time: The date/time of the run.
            result_format: One of several supported formatting directives for expectation validation results
            kwargs: Additional kwargs

        Returns:
            Checkpoint result.
        """
        batches_to_validate = self._get_batches_to_validate(self.batches)

        if (
            self.validation_operator_name
            and self.data_context.validation_operators.get(
                self.validation_operator_name
            )
        ):
            results = self.data_context.run_validation_operator(
                self.validation_operator_name,
                assets_to_validate=batches_to_validate,
                run_id=run_id,
                evaluation_parameters=evaluation_parameters,
                run_name=run_name,
                run_time=run_time,
                result_format=result_format,
                **kwargs,
            )
        else:
            if self.validation_operator_name:
                logger.warning(
                    f'Could not find Validation Operator "{self.validation_operator_name}" when '
                    f'running Checkpoint "{self.name}". Using default action_list_operator.'
                )

            results = self._run_default_validation_operator(
                assets_to_validate=batches_to_validate,
                run_id=run_id,
                evaluation_parameters=evaluation_parameters,
                run_name=run_name,
                run_time=run_time,
                result_format=result_format,
            )

        return results

    def _get_batches_to_validate(self, batches):
        batches_to_validate = []
        for batch in batches:

            batch_kwargs = batch["batch_kwargs"]
            suites = batch["expectation_suite_names"]

            if not suites:
                raise Exception(
                    f"""A batch has no suites associated with it. At least one suite is required.
    - Batch: {json.dumps(batch_kwargs)}
    - Please add at least one suite to Checkpoint {self.name}
"""
                )

            for suite_name in batch["expectation_suite_names"]:
                suite = self.data_context.get_expectation_suite(suite_name)
                batch = self.data_context.get_batch(batch_kwargs, suite)

                batches_to_validate.append(batch)

        return batches_to_validate


@public_api
@deprecated_argument(argument_name="validation_operator_name", version="0.14.0")
@deprecated_argument(argument_name="batches", version="0.14.0")
@new_argument(
    argument_name="ge_cloud_id", version="0.13.33", message="Used in cloud deployments."
)
@new_argument(
    argument_name="expectation_suite_ge_cloud_id",
    version="0.13.33",
    message="Used in cloud deployments.",
)
class SimpleCheckpoint(Checkpoint):
    """A SimpleCheckpoint provides a means to simplify the process of specifying a Checkpoint configuration.

    It provides a basic set of actions - store Validation Result, store Evaluation Parameters, update Data Docs,
    and optionally, send a Slack notification - allowing you to omit an action_list from your configuration and at runtime.

    Args:
        name: user-selected Checkpoint name (e.g. `staging_tables`).
        data_context: Data context that is associated with the current checkpoint.
        config_version: version number of the checkpoint configuration.
        template_name: the name of another checkpoint to use as a base template.
        run_name_template: a template to create run names, using environment variables and datetime-template syntax (e.g. `%Y-%M-staging-$MY_ENV_VAR`).
        expectation_suite_name: expectation suite associated with checkpoint.
        batch_request: batch request describing the batch of data to validate.
        action_list: a list of actions to perform after each batch is validated.
        evaluation_parameters: evaluation parameters to use in generating this checkpoint.
        runtime_configuration: runtime configuration to pass into the validator's runtime configuration (e.g. `result_format`).
        validations: validations to be executed as part of checkpoint.
        profilers: profilers to use in generating this checkpoint.
        validation_operator_name: list of validation operators configured by the checkpoint.
        batches: list of batches for validation by checkpoint.
        ge_cloud_id: Great Expectations Cloud id for this checkpoint.
        site_names: a list of Data Docs site names to update as part of the update Data Docs action - defaults to `all`.
        slack_webhook:  if provided, an action will be added that sends a Slack notification to the provided webhook.
        notify_on: used to define when a notification is fired, according to validation result outcome - `all`, `failure`, or `success`. Defaults to `all`.
        notify_with: a list of Data Docs site names for which to include a URL in any notifications - defaults to `all`.
        expectation_suite_ge_cloud_id: Great Expectations Cloud id associated with expectation suite.
        kwargs: additional keyword arguments.
    """

    _configurator_class = SimpleCheckpointConfigurator

    # noinspection PyUnusedLocal
    def __init__(
        self,
        name: str,
        data_context,
        config_version: Optional[Union[int, float]] = None,
        template_name: Optional[str] = None,
        run_name_template: Optional[str] = None,
        expectation_suite_name: Optional[str] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        action_list: Optional[List[dict]] = None,
        evaluation_parameters: Optional[dict] = None,
        runtime_configuration: Optional[dict] = None,
        validations: Optional[List[dict]] = None,
        profilers: Optional[List[dict]] = None,
        ge_cloud_id: Optional[UUID] = None,
        # the following four arguments are used by SimpleCheckpointConfigurator
        site_names: Union[str, List[str]] = "all",
        slack_webhook: Optional[str] = None,
        notify_on: str = "all",
        notify_with: Union[str, List[str]] = "all",
        expectation_suite_ge_cloud_id: Optional[str] = None,
        **kwargs,
    ) -> None:

        checkpoint_config: CheckpointConfig = self._configurator_class(
            name=name,
            data_context=data_context,
            config_version=config_version,
            template_name=template_name,
            run_name_template=run_name_template,
            expectation_suite_name=expectation_suite_name,
            batch_request=batch_request,
            action_list=action_list,
            evaluation_parameters=evaluation_parameters,
            runtime_configuration=runtime_configuration,
            validations=validations,
            profilers=profilers,
            site_names=site_names,
            slack_webhook=slack_webhook,
            notify_on=notify_on,
            notify_with=notify_with,
            ge_cloud_id=ge_cloud_id,
            expectation_suite_ge_cloud_id=expectation_suite_ge_cloud_id,
        ).build()

        super().__init__(
            name=checkpoint_config.name,
            data_context=data_context,
            config_version=checkpoint_config.config_version,
            template_name=checkpoint_config.template_name,
            run_name_template=checkpoint_config.run_name_template,
            expectation_suite_name=checkpoint_config.expectation_suite_name,
            batch_request=batch_request,
            action_list=checkpoint_config.action_list,
            evaluation_parameters=checkpoint_config.evaluation_parameters,
            runtime_configuration=checkpoint_config.runtime_configuration,
            validations=validations,
            profilers=checkpoint_config.profilers,
            ge_cloud_id=checkpoint_config.ge_cloud_id,
            expectation_suite_ge_cloud_id=checkpoint_config.expectation_suite_ge_cloud_id,
        )

    @public_api
    @new_argument(
        argument_name="expectation_suite_ge_cloud_id",
        version="0.13.33",
        message="Used in cloud deployments.",
    )
    def run(
        self,
        template_name: Optional[str] = None,
        run_name_template: Optional[str] = None,
        expectation_suite_name: Optional[str] = None,
        batch_request: Optional[Union[BatchRequestBase, dict]] = None,
        action_list: Optional[List[dict]] = None,
        evaluation_parameters: Optional[dict] = None,
        runtime_configuration: Optional[dict] = None,
        validations: Optional[List[dict]] = None,
        profilers: Optional[List[dict]] = None,
        run_id: Optional[Union[str, RunIdentifier]] = None,
        run_name: Optional[str] = None,
        run_time: Optional[Union[str, datetime.datetime]] = None,
        result_format: Optional[str] = None,
        # the following four arguments are specific to SimpleCheckpoint
        site_names: Union[str, List[str]] = "all",
        slack_webhook: Optional[str] = None,
        notify_on: str = "all",
        notify_with: Union[str, List[str]] = "all",
        expectation_suite_ge_cloud_id: Optional[str] = None,
    ) -> CheckpointResult:
        """Validate against the current SimpleCheckpoint.

        Arguments allow for override of the current SimpleCheckpoint configuration.

        Args:
            template_name: The name of another checkpoint to use as a base template.
            run_name_template: A template to create run names, using environment
                variables and datetime-template syntax (e.g. "%Y-%M-staging-$MY_ENV_VAR").
            expectation_suite_name: Expectation suite associated with checkpoint.
            batch_request: Batch request describing the batch of data to validate.
            action_list: A list of actions to perform after each batch is validated.
            evaluation_parameters: Evaluation parameters to use in generating this checkpoint.
            runtime_configuration: Runtime configuration to pass into the validator's runtime configuration
                (e.g. `result_format`).
            validations: Validations to be executed as part of checkpoint.
            profilers: Profilers to use in generating this checkpoint.
            run_id: The run_id for the validation; if None, a default value will be used.
            run_name: The run_name for the validation; if None, a default value will be used.
            run_time: The date/time of the run.
            result_format: One of several supported formatting directives for expectation validation results
            site_names: a list of Data Docs site names to update as part of the update Data Docs action - defaults to `all`.
            slack_webhook:  if provided, an action will be added that sends a Slack notification to the provided webhook.
            notify_on: used to define when a notification is fired, according to validation result outcome - `all`, `failure`, or `success`. Defaults to `all`.
            notify_with: a list of Data Docs site names for which to include a URL in any notifications - defaults to `all`.
            expectation_suite_ge_cloud_id: Great Expectations Cloud id for the expectation suite

        Returns:
            CheckpointResult
        """
        new_baseline_config = None

        # if any SimpleCheckpoint-specific kwargs are passed, generate a new baseline config using configurator,
        # passing only action_list, since this is the only config key that would be affected by the
        # SimpleCheckpoint-specific kwargs
        if any((site_names, slack_webhook, notify_on, notify_with)):
            new_baseline_config = self._configurator_class(
                name=self.name,
                data_context=self.data_context,
                action_list=action_list,
                site_names=site_names,
                slack_webhook=slack_webhook,
                notify_on=notify_on,
                notify_with=notify_with,
            ).build()

        return super().run(
            template_name=template_name,
            run_name_template=run_name_template,
            expectation_suite_name=expectation_suite_name,
            batch_request=batch_request,
            action_list=new_baseline_config.action_list
            if new_baseline_config
            else action_list,
            evaluation_parameters=evaluation_parameters,
            runtime_configuration=runtime_configuration,
            validations=validations,
            profilers=profilers,
            run_id=run_id,
            run_name=run_name,
            run_time=run_time,
            result_format=result_format,
            expectation_suite_ge_cloud_id=expectation_suite_ge_cloud_id,
        )



import warnings
from typing import List, Optional, Union

from great_expectations.core._docs_decorators import deprecated_method_or_class
from great_expectations.render import (
    AtomicDiagnosticRendererType,
    AtomicPrescriptiveRendererType,
)
from great_expectations.render import CollapseContent as CollapseContentRender
from great_expectations.render import (
    RenderedAtomicContent as RenderedAtomicContentRender,
)
from great_expectations.render import (
    RenderedAtomicContentSchema as RenderedAtomicContentSchemaRender,
)
from great_expectations.render import RenderedAtomicValue as RenderedAtomicValueRender
from great_expectations.render import (
    RenderedAtomicValueGraph as RenderedAtomicValueGraphRender,
)
from great_expectations.render import (
    RenderedAtomicValueSchema as RenderedAtomicValueSchemaRender,
)
from great_expectations.render import (
    RenderedBootstrapTableContent as RenderedBootstrapTableContentRender,
)
from great_expectations.render import (
    RenderedBulletListContent as RenderedBulletListContentRender,
)
from great_expectations.render import (
    RenderedComponentContent as RenderedComponentContentRender,
)
from great_expectations.render import RenderedContent as RenderedContentRender
from great_expectations.render import (
    RenderedContentBlockContainer as RenderedContentBlockContainerRender,
)
from great_expectations.render import (
    RenderedDocumentContent as RenderedDocumentContentRender,
)
from great_expectations.render import RenderedGraphContent as RenderedGraphContentRender
from great_expectations.render import (
    RenderedHeaderContent as RenderedHeaderContentRender,
)
from great_expectations.render import (
    RenderedMarkdownContent as RenderedMarkdownContentRender,
)
from great_expectations.render import (
    RenderedSectionContent as RenderedSectionContentRender,
)
from great_expectations.render import (
    RenderedStringTemplateContent as RenderedStringTemplateContentRender,
)
from great_expectations.render import RenderedTableContent as RenderedTableContentRender
from great_expectations.render import RenderedTabsContent as RenderedTabsContentRender
from great_expectations.render import TextContent as TextContentRender
from great_expectations.render import ValueListContent as ValueListContentRender
from great_expectations.render.renderer_configuration import (
    RendererTableValue as RendererTableValueRender,  # noqa: TCH001
)


# TODO: Remove this entire module for release 0.18.0
def _get_deprecation_warning_message(classname: str) -> str:
    return (
        f"Importing the class {classname} from great_expectations.render.types is deprecated as of v0.15.32 "
        f"in v0.18. Please import class {classname} from great_expectations.render."
    )


@deprecated_method_or_class(
    version="0.15.32",
    message="Importing the class CollapseContent from "
    "great_expectations.render.types is deprecated as of v0.15.32 "
    "in v0.18. Please import class CollapseContent "
    "from great_expectations.render.",
)
class CollapseContent(CollapseContentRender):
    # deprecated-v0.15.32
    """CollapseContent is RenderedComponentContent that can be collapsed.

    Args:
        collapse: The content to be collapsed. If a list is provided, it can recursively contain RenderedContent.
        collpase_toggle_link: The toggle link for this CollapseContent.
        header: The header for this content block.
        subheader: The subheader for this content block.
        styling: A dictionary containing styling information.
        content_block_type: The type of content block.
        inline_link: Whether to include a link inline.
    """

    def __init__(
        self,
        collapse: Union[RenderedContentRender, list],
        collapse_toggle_link: Optional[Union[RenderedContentRender, dict]] = None,
        header: Optional[Union[RenderedContentRender, dict]] = None,
        subheader: Optional[Union[RenderedContentRender, dict]] = None,
        styling: Optional[dict] = None,
        content_block_type: str = "collapse",
        inline_link: bool = False,
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            collapse=collapse,
            collapse_toggle_link=collapse_toggle_link,
            header=header,
            subheader=subheader,
            styling=styling,
            content_block_type=content_block_type,
            inline_link=inline_link,
        )


class RenderedAtomicContent(RenderedAtomicContentRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        name: Union[str, AtomicDiagnosticRendererType, AtomicPrescriptiveRendererType],
        value: RenderedAtomicValueRender,
        value_type: Optional[str] = None,
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            name=name,
            value=value,
            value_type=value_type,
        )


class RenderedAtomicContentSchema(RenderedAtomicContentSchemaRender):
    # deprecated-v0.15.32
    def __new__(cls):
        warnings.warn(
            _get_deprecation_warning_message(classname=cls.__name__),
            DeprecationWarning,
        )


class RenderedAtomicValue(RenderedAtomicValueRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        schema: Optional[dict] = None,
        header: Optional[RenderedAtomicValueRender] = None,
        template: Optional[str] = None,
        params: Optional[dict] = None,
        header_row: Optional[List[RendererTableValueRender]] = None,
        table: Optional[List[List[RendererTableValueRender]]] = None,
        graph: Optional[dict] = None,
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            schema=schema,
            header=header,
            template=template,
            params=params,
            header_row=header_row,
            table=table,
            graph=graph,
        )


class RenderedAtomicValueGraph(RenderedAtomicValueGraphRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        graph: Optional[dict] = None,
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            graph=graph,
        )


class RenderedAtomicValueSchema(RenderedAtomicValueSchemaRender):
    # deprecated-v0.15.32
    def __new__(cls):
        warnings.warn(
            _get_deprecation_warning_message(classname=cls.__name__),
            DeprecationWarning,
        )


class RenderedBootstrapTableContent(RenderedBootstrapTableContentRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        table_data,
        table_columns,
        title_row=None,
        table_options=None,
        header=None,
        subheader=None,
        styling=None,
        content_block_type="bootstrap_table",
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            table_data=table_data,
            table_columns=table_columns,
            title_row=title_row,
            table_options=table_options,
            header=header,
            subheader=subheader,
            styling=styling,
            content_block_type=content_block_type,
        )


class RenderedBulletListContent(RenderedBulletListContentRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        bullet_list,
        header=None,
        subheader=None,
        styling=None,
        content_block_type="bullet_list",
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            bullet_list=bullet_list,
            header=header,
            subheader=subheader,
            styling=styling,
            content_block_type=content_block_type,
        )


class RenderedComponentContent(RenderedComponentContentRender):
    # deprecated-v0.15.32
    def __init__(self, content_block_type, styling=None):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            content_block_type=content_block_type,
            styling=styling,
        )


class RenderedContent(RenderedContentRender):
    # deprecated-v0.15.32
    def __new__(cls):
        warnings.warn(
            _get_deprecation_warning_message(classname=cls.__name__),
            DeprecationWarning,
        )


class RenderedContentBlockContainer(RenderedContentBlockContainerRender):
    # deprecated-v0.15.32
    def __init__(
        self, content_blocks, styling=None, content_block_type="content_block_container"
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            content_blocks=content_blocks,
            styling=styling,
            content_block_type=content_block_type,
        )


class RenderedDocumentContent(RenderedDocumentContentRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        sections,
        data_asset_name=None,
        full_data_asset_identifier=None,
        renderer_type=None,
        page_title=None,
        utm_medium=None,
        cta_footer=None,
        expectation_suite_name=None,
        batch_kwargs=None,
        batch_spec=None,
        ge_cloud_id=None,
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            sections=sections,
            data_asset_name=data_asset_name,
            full_data_asset_identifier=full_data_asset_identifier,
            renderer_type=renderer_type,
            page_title=page_title,
            utm_medium=utm_medium,
            cta_footer=cta_footer,
            expectation_suite_name=expectation_suite_name,
            batch_kwargs=batch_kwargs,
            batch_spec=batch_spec,
            ge_cloud_id=ge_cloud_id,
        )


class RenderedGraphContent(RenderedGraphContentRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        graph,
        header=None,
        subheader=None,
        styling=None,
        content_block_type="graph",
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            graph=graph,
            header=header,
            subheader=subheader,
            styling=styling,
            content_block_type=content_block_type,
        )


class RenderedHeaderContent(RenderedHeaderContentRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        header,
        subheader=None,
        header_row=None,
        styling=None,
        content_block_type="header",
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            header=header,
            subheader=subheader,
            header_row=header_row,
            styling=styling,
            content_block_type=content_block_type,
        )


class RenderedMarkdownContent(RenderedMarkdownContentRender):
    # deprecated-v0.15.32
    def __init__(self, markdown, styling=None, content_block_type="markdown"):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            markdown=markdown,
            styling=styling,
            content_block_type=content_block_type,
        )


class RenderedSectionContent(RenderedSectionContentRender):
    # deprecated-v0.15.32
    def __init__(self, content_blocks, section_name=None):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            content_blocks=content_blocks,
            section_name=section_name,
        )


@deprecated_method_or_class(
    version="0.15.32",
    message="Importing the class RenderedStringTemplateContent from "
    "great_expectations.render.types is deprecated as of v0.15.32 "
    "in v0.18. Please import class RenderedStringTemplateContent "
    "from great_expectations.render.",
)
class RenderedStringTemplateContent(RenderedStringTemplateContentRender):
    # deprecated-v0.15.32
    """RenderedStringTemplateContent is RenderedComponentContent that represents a templated string.

    Args:
        string_template: A dictionary containing:

            template: The string to perform substitution on. Variables are denoted with a preceeding $.

            params: A dictionary with keys that match variable names and values which will be substituted.

            styling: A dictionary containing styling information.
        styling: A dictionary containing styling information.
        content_block_type: The type of content block.
    """

    def __init__(
        self, string_template: dict, styling=None, content_block_type="string_template"
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            string_template=string_template,
            styling=styling,
            content_block_type=content_block_type,
        )


@deprecated_method_or_class(
    version="0.15.32",
    message="Importing the class RenderedTableContent from "
    "great_expectations.render.types is deprecated as of v0.15.32 "
    "in v0.18. Please import class RenderedTableContent "
    "from great_expectations.render.",
)
class RenderedTableContent(RenderedTableContentRender):
    # deprecated-v0.15.32
    """RenderedTableContent is RenderedComponentContent that is a table.

    Args:
        table: The table to be rendered.
        header: The header for this content block.
        subheader: The subheader for this content block.
        header_row: The header row for the table.
        styling: A dictionary containing styling information.
        content_block_type: The type of content block.
        table_options: The options that can be set for the table.

            search: A boolean indicating whether to include search with the table.

            icon-size: The size of the icons in the table. One of "sm", "md", or "lg".
        header_row_options: The options that can be set for the header_row. A dictionary with the keys being the column
            name and the values being a dictionary with the following form:

            sortable: A boolean indicating whether the column is sortable.
    """

    def __init__(
        self,
        table,
        header=None,
        subheader=None,
        header_row=None,
        styling=None,
        content_block_type="table",
        table_options=None,
        header_row_options=None,
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            table=table,
            header=header,
            subheader=subheader,
            header_row=header_row,
            styling=styling,
            content_block_type=content_block_type,
            table_options=table_options,
            header_row_options=header_row_options,
        )


class RenderedTabsContent(RenderedTabsContentRender):
    # deprecated-v0.15.32
    def __init__(
        self, tabs, header=None, subheader=None, styling=None, content_block_type="tabs"
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            tabs=tabs,
            header=header,
            subheader=subheader,
            styling=styling,
            content_block_type=content_block_type,
        )


class TextContent(TextContentRender):
    # deprecated-v0.15.32
    def __init__(
        self, text, header=None, subheader=None, styling=None, content_block_type="text"
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            text=text,
            header=header,
            subheader=subheader,
            styling=styling,
            content_block_type=content_block_type,
        )


class ValueListContent(ValueListContentRender):
    # deprecated-v0.15.32
    def __init__(
        self,
        value_list,
        header=None,
        subheader=None,
        styling=None,
        content_block_type="value_list",
    ):
        warnings.warn(
            _get_deprecation_warning_message(classname=self.__class__.__name__),
            DeprecationWarning,
        )
        super().__init__(
            value_list=value_list,
            header=header,
            subheader=subheader,
            styling=styling,
            content_block_type=content_block_type,
        )



import copy
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Set, TypeVar, Union

from pyparsing import (
    Literal,
    ParseException,
    ParseResults,
    Suppress,
    Word,
    ZeroOrMore,
    alphanums,
    alphas,
    nums,
)
from typing_extensions import Final

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.domain import Domain  # noqa: TCH001
from great_expectations.core.util import convert_to_json_serializable
from great_expectations.types import SerializableDictDot, SerializableDotDict

FULLY_QUALIFIED_PARAMETER_NAME_DELIMITER_CHARACTER: Final[str] = "$"

FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER: Final[str] = "."

DOMAIN_KWARGS_PARAMETER_NAME: Final[str] = "domain_kwargs"
DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME: Final[
    str
] = f"{FULLY_QUALIFIED_PARAMETER_NAME_DELIMITER_CHARACTER}domain{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{DOMAIN_KWARGS_PARAMETER_NAME}"

PARAMETER_NAME_ROOT_FOR_VARIABLES: Final[str] = "variables"
VARIABLES_PREFIX: Final[
    str
] = f"{FULLY_QUALIFIED_PARAMETER_NAME_DELIMITER_CHARACTER}{PARAMETER_NAME_ROOT_FOR_VARIABLES}"
VARIABLES_KEY: Final[
    str
] = f"{VARIABLES_PREFIX}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}"

PARAMETER_NAME_ROOT_FOR_PARAMETERS: Final[str] = "parameter"
PARAMETER_PREFIX: Final[
    str
] = f"{FULLY_QUALIFIED_PARAMETER_NAME_DELIMITER_CHARACTER}{PARAMETER_NAME_ROOT_FOR_PARAMETERS}"
PARAMETER_KEY: Final[
    str
] = f"{PARAMETER_PREFIX}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}"

RAW_SUFFIX: Final[str] = "raw"
RAW_PARAMETER_KEY: Final[
    str
] = f"{PARAMETER_KEY}{RAW_SUFFIX}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}"

FULLY_QUALIFIED_PARAMETER_NAME_ATTRIBUTED_VALUE_KEY: Final[str] = "attributed_value"
FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY: Final[str] = "details"
FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY: Final[str] = "value"

RESERVED_TERMINAL_LITERALS: Final[Set[str]] = {
    FULLY_QUALIFIED_PARAMETER_NAME_ATTRIBUTED_VALUE_KEY,
    FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY,
    FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY,
}

attribute_naming_pattern = Word(alphas, alphanums + "_.") + ZeroOrMore(
    (
        (
            Suppress(Literal('["'))
            + Word(alphas, alphanums + "_.")
            + Suppress(Literal('"]'))
        )
        ^ (
            Suppress(Literal("['"))
            + Word(alphas, alphanums + "_.")
            + Suppress(Literal("']"))
        )
    )
    ^ (
        Suppress(Literal("["))
        + Word(nums + "-").setParseAction(
            lambda s, l, t: [int(t[0])]  # noqa: E741 # ambiguous var name
        )
        + Suppress(Literal("]"))
    )
)


T = TypeVar("T")


class ParameterAttributeNameParserError(gx_exceptions.GreatExpectationsError):
    pass


def _parse_attribute_naming_pattern(name: str) -> ParseResults:
    """
    Using grammer defined by "attribute_naming_pattern", parses collection (list, dictionary) access syntax:
    List: variable[index: int]
    Dictionary: variable[key: str]
    Nested List/Dictionary: variable[index_0: int][key_0: str][index_1: int][key_1: str][key_2: str][index_2: int]...

    Applicability: To be used as part of configuration (e.g., YAML-based files or text strings).
    Extendability: Readily extensible to include "slice" and other standard accessors (as long as no dynamic elements).
    """

    try:
        return attribute_naming_pattern.parseString(name)
    except ParseException:
        raise ParameterAttributeNameParserError(
            f'Unable to parse Parameter Attribute Name: "{name}".'
        )


def is_fully_qualified_parameter_name_literal_string_format(
    fully_qualified_parameter_name: str,
) -> bool:
    return fully_qualified_parameter_name.startswith(
        f"{FULLY_QUALIFIED_PARAMETER_NAME_DELIMITER_CHARACTER}"
    )


def validate_fully_qualified_parameter_name(
    fully_qualified_parameter_name: str,
) -> None:
    if not is_fully_qualified_parameter_name_literal_string_format(
        fully_qualified_parameter_name=fully_qualified_parameter_name
    ):
        raise gx_exceptions.ProfilerExecutionError(
            message=f"""Unable to get value for parameter name "{fully_qualified_parameter_name}" -- parameter \
names must start with {FULLY_QUALIFIED_PARAMETER_NAME_DELIMITER_CHARACTER} (e.g., "{FULLY_QUALIFIED_PARAMETER_NAME_DELIMITER_CHARACTER}{fully_qualified_parameter_name}").
"""
        )


class ParameterNode(SerializableDotDict):
    """
    ParameterNode is a node of a tree structure.

    The tree is implemented as a nested dictionary that also supports the "dot" notation at every level of hierarchy.
    Together, these design aspects allow the entire tree to be converted into a JSON object for external compatibility.

    Since the descendant nodes (i.e., sub-dictionaries) are of the same type as their parent node, then each descendant
    node is also a tree (or a sub-tree).  Each node can support the combination of attribute name-value pairs
    representing values and details containing helpful information regarding how these values were obtained (tolerances,
    explanations, etc.).  By convention, the "value" key corresponds the parameter value, while the "details" key
    corresponds the auxiliary details.  These details can be used to set the "meta" key of the ExpectationConfiguration.

    See the ParameterContainer documentation for examples of different parameter naming structures supported.

    Even though, typically, only the leaf nodes (characterized by having no keys of "ParameterNode" type) store
    parameter values and details, intermediate nodes may also have these properties.  This is important for supporting
    the situations where multiple long fully-qualified parameter names have overlapping intermediate parts (see below).
    """

    def to_dict(self) -> dict:
        return convert_parameter_node_to_dictionary(source=dict(self))  # type: ignore[return-value] # could be None

    def to_json_dict(self) -> dict:
        return convert_to_json_serializable(data=self.to_dict())


@dataclass
class ParameterContainer(SerializableDictDot):
    """
    ParameterContainer holds root nodes of tree structures, corresponding to fully-qualified parameter names.

    While all parameter names begin with the dollar sign character ("$"), a fully-qualified parameter name is a string,
    whose parts are delimited by the period character (".").

    As an example, suppose that the value of the attribute "max_num_conversion_attempts" is needed for certain
    processing operations.  However, there could be several attributes with this name, albeit in different contexts:
    $parameter.date_strings.tolerances.max_num_conversion_attempts
    $parameter.character.encodings.tolerances.max_num_conversion_attempts
    The fully-qualified parameter names disambiguate the same "leaf" attribute names occurring in multiple contexts.
    In the present example, the use of fully-qualified parameter names makes it clear that in one context,
    "max_num_conversion_attempts" refers to the operations on date/time, while in the other -- it applies to characters.

    $variables.false_positive_threshold
    $parameter.date_strings.yyyy_mm_dd_hh_mm_ss_tz_date_format.value
    $parameter.date_strings.yyyy_mm_dd_hh_mm_ss_tz_date_format.details
    $parameter.date_strings.yyyy_mm_dd_date_format.value
    $parameter.date_strings.yyyy_mm_dd_date_format.details
    $parameter.date_strings.mm_yyyy_dd_hh_mm_ss_tz_date_format.value
    $parameter.date_strings.mm_yyyy_dd_hh_mm_ss_tz_date_format.details
    $parameter.date_strings.mm_yyyy_dd_date_format.value
    $parameter.date_strings.mm_yyyy_dd_date_format.details
    $parameter.date_strings.tolerances.max_abs_error_time_milliseconds
    $parameter.date_strings.tolerances.max_num_conversion_attempts
    $parameter.tolerances.mostly
    $parameter.tolerances.financial.usd
    $mean
    $parameter.monthly_taxi_fairs.mean_values.value[0]
    $parameter.monthly_taxi_fairs.mean_values.value[1]
    $parameter.monthly_taxi_fairs.mean_values.value[2]
    $parameter.monthly_taxi_fairs.mean_values.value[3]
    $parameter.monthly_taxi_fairs.mean_values.details
    $parameter.daily_taxi_fairs.mean_values.value["friday"]
    $parameter.daily_taxi_fairs.mean_values.value["saturday"]
    $parameter.daily_taxi_fairs.mean_values.value["sunday"]
    $parameter.daily_taxi_fairs.mean_values.value["monday"]
    $parameter.daily_taxi_fairs.mean_values.details
    $parameter.weekly_taxi_fairs.mean_values.value[1]['friday']
    $parameter.weekly_taxi_fairs.mean_values.value[18]['saturday']
    $parameter.weekly_taxi_fairs.mean_values.value[20]['sunday']
    $parameter.weekly_taxi_fairs.mean_values.value[21]['monday']
    $parameter.weekly_taxi_fairs.mean_values.details
    $custom.lang.character_encodings

    The reason that ParameterContainer is needed is that each ParameterNode can point only to one tree structure,
    characterized by having a specific root-level ParameterNode object.  A root-level ParameterNode object corresponds
    to a set of fully-qualified parameter names that have the same first part (e.g., "parameter").  However, a Domain
    may utilize fully-qualified parameter names that have multiple first parts (i.e., from different "name spaces").
    The ParameterContainer maintains a dictionary that holds references to root-level ParameterNode objects for all
    parameter "name spaces" applicable to the given Domain (where the first part of all fully-qualified parameter names
    within the same "name space" serves as the dictionary key, and the root-level ParameterNode objects are the values).
    """

    parameter_nodes: Optional[Dict[str, ParameterNode]] = None

    def set_parameter_node(
        self, parameter_name_root: str, parameter_node: ParameterNode
    ) -> None:
        if self.parameter_nodes is None:
            self.parameter_nodes = {}

        self.parameter_nodes[parameter_name_root] = parameter_node

    def get_parameter_node(self, parameter_name_root: str) -> Optional[ParameterNode]:
        if self.parameter_nodes is None:
            return None

        return self.parameter_nodes.get(parameter_name_root)

    def to_dict(self) -> dict:
        return asdict(self)

    def to_json_dict(self) -> dict:
        return convert_to_json_serializable(data=self.to_dict())


def deep_convert_properties_iterable_to_parameter_node(
    source: Union[T, dict]
) -> Union[T, ParameterNode]:
    if isinstance(source, dict):
        return _deep_convert_properties_iterable_to_parameter_node(
            source=ParameterNode(source)
        )

    # Must allow for non-dictionary source types, since their internal nested structures may contain dictionaries.
    if isinstance(source, (list, set, tuple)):
        data_type: type = type(source)

        element: Any
        return data_type(
            [
                deep_convert_properties_iterable_to_parameter_node(source=element)
                for element in source
            ]
        )

    return source


def _deep_convert_properties_iterable_to_parameter_node(source: dict) -> ParameterNode:
    key: str
    value: Any
    for key, value in source.items():
        if isinstance(value, dict):
            source[key] = _deep_convert_properties_iterable_to_parameter_node(
                source=value
            )
        elif isinstance(value, (list, set, tuple)):
            data_type: type = type(value)

            element: Any
            source[key] = data_type(
                [
                    deep_convert_properties_iterable_to_parameter_node(source=element)
                    for element in value
                ]
            )

    return ParameterNode(source)


def convert_parameter_node_to_dictionary(
    source: Optional[Any],
) -> Optional[dict]:
    if source is None:
        return None

    if isinstance(source, ParameterNode):
        source = source.to_dict()

        key: str
        value: Any
        for key, value in source.items():
            source[key] = convert_parameter_node_to_dictionary(source=value)

    return source


def build_parameter_container_for_variables(
    variables_configs: Dict[str, Any]
) -> ParameterContainer:
    """
    Build a ParameterContainer for all of the profiler config variables passed as key value pairs
    Args:
        variables_configs: Variable key: value pairs e.g. {"variable_name": variable_value, ...}

    Returns:
        ParameterContainer containing all variables
    """
    parameter_node = ParameterNode()
    key: str
    value: Any
    for key, value in variables_configs.items():
        key_parts: List[str] = [PARAMETER_NAME_ROOT_FOR_VARIABLES] + key.split(
            FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER
        )
        _build_parameter_node_tree_for_one_parameter(parameter_node, key_parts, value)

    # We only want to set the ParameterContainer key PARAMETER_NAME_ROOT_FOR_VARIABLES if
    # parameter_node is non-empty since there is downstream logic that depends on this.
    return (
        ParameterContainer(
            parameter_nodes={PARAMETER_NAME_ROOT_FOR_VARIABLES: parameter_node}
        )
        if parameter_node
        else ParameterContainer()
    )


def build_parameter_container(
    parameter_container: ParameterContainer,
    parameter_values: Dict[str, Any],
) -> None:
    """
    Builds the ParameterNode trees, corresponding to the fully_qualified_parameter_name first-level keys.

    :param parameter_container initialized ParameterContainer for all ParameterNode trees
    :param parameter_values
    Example of the name-value structure for building parameters (matching the type hint in the method signature):
    {
        "$parameter.date_strings.tolerances.max_abs_error_time_milliseconds.value": 100, # Actual value can of Any type.
        # The "details" dictionary is Optional.
        "$parameter.date_strings.tolerances.max_abs_error_time_milliseconds.details": {
            "max_abs_error_time_milliseconds": {
                "confidence": {  # Arbitrary dictionary key
                    "success_ratio": 1.0,  # Arbitrary entries
                    "comment": "matched template",  # Arbitrary entries
                }
            },
        },
        # While highly recommended, the use of ".value" and ".details" keys is conventional (it is not enforced).
        "$parameter.tolerances.mostly": 9.0e-1,  # The key here does not end on ".value" and no ".details" is provided.
        ...
    }
    :return parameter_container holds the dictionary of ParameterNode objects corresponding to roots of parameter names

    This function loops through the supplied pairs of fully-qualified parameter names and their corresponding values
    (and any "details") and builds the tree under a single root-level ParameterNode object for a "name space".
    In particular, if any ParameterNode object in the tree (starting with the root-level ParameterNode object) already
    exists, it is reused; in other words, ParameterNode objects are unique per part of fully-qualified parameter names.
    """
    parameter_node: Optional[ParameterNode]
    fully_qualified_parameter_name: str
    parameter_value: Any
    fully_qualified_parameter_name_as_list: List[str]
    parameter_name_root: str
    for (
        fully_qualified_parameter_name,
        parameter_value,
    ) in parameter_values.items():
        validate_fully_qualified_parameter_name(
            fully_qualified_parameter_name=fully_qualified_parameter_name
        )
        fully_qualified_parameter_name_as_list = fully_qualified_parameter_name[
            1:
        ].split(FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER)
        parameter_name_root = fully_qualified_parameter_name_as_list[0]
        parameter_node = parameter_container.get_parameter_node(
            parameter_name_root=parameter_name_root
        )
        if parameter_node is None:
            parameter_node = ParameterNode({})
            parameter_container.set_parameter_node(
                parameter_name_root=parameter_name_root, parameter_node=parameter_node
            )

        _build_parameter_node_tree_for_one_parameter(
            parameter_node=parameter_node,
            parameter_name_as_list=fully_qualified_parameter_name_as_list,
            parameter_value=parameter_value,
        )


def _build_parameter_node_tree_for_one_parameter(
    parameter_node: ParameterNode,
    parameter_name_as_list: List[str],
    parameter_value: Any,
) -> None:
    """
    Builds a tree of ParameterNode objects.

    parameter_name_as_list is a list of property names which are used to access into parameter_node. If the property
    doesn't exist, it is created. The parameter_value is assigned to the lowest most property.
    For example if parameter_name_as_list is ["a", "b", "c"] and parameter_value is "value" then parameter_node is
    modified in place so that:

    parameter_node.a.b.c = parameter_value

    Args:
        parameter_node: root-level ParameterNode for the sub-tree, characterized by the first parameter name in list
        parameter_name_as_list: list of parts of a fully-qualified parameter name of sub-tree (or sub "name space")
        parameter_value: value pertaining to the last part of the fully-qualified parameter name ("leaf node")
    """
    node: ParameterNode = parameter_node
    parameter_name: str
    for parameter_name in parameter_name_as_list[:-1]:
        # This conditional is functionally equivalent to `node = node.setdefault(parameter_name, ParameterNode({})).`
        # However, setdefault always evaluates its second argument which is much slower in this hot code path.
        if parameter_name in node:
            node = node[parameter_name]
        else:
            node[parameter_name] = ParameterNode({})
            node = node[parameter_name]

    node[
        parameter_name_as_list[-1]
    ] = deep_convert_properties_iterable_to_parameter_node(parameter_value)


def get_parameter_value_by_fully_qualified_parameter_name(
    fully_qualified_parameter_name: str,
    domain: Optional[Domain] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> Optional[Union[Any, ParameterNode]]:
    """
    Get the parameter value from the current "rule state" using the fully-qualified parameter name.
    A fully-qualified parameter name must be a dot-delimited string, or the name of a parameter (without the dots).
    Args
        :param fully_qualified_parameter_name: str -- A dot-separated string key starting with $ for fetching parameters
        :param domain: Domain -- current Domain of interest
        :param variables
        :param parameters
    :return: Optional[Union[Any, ParameterNode]] object corresponding to the last part of the fully-qualified parameter
    name supplied as argument -- a value (of type "Any") or a ParameterNode object (containing the sub-tree structure).
    """
    validate_fully_qualified_parameter_name(
        fully_qualified_parameter_name=fully_qualified_parameter_name
    )

    # Using "__getitem__" (bracket) notation instead of "__getattr__" (dot) notation in order to insure the
    # compatibility of field names (e.g., "domain_kwargs") with user-facing syntax (as governed by the value of the
    # DOMAIN_KWARGS_PARAMETER_NAME constant, which may change, requiring the same change to the field name).
    if fully_qualified_parameter_name == DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME:
        if domain:
            # Supports the "$domain.domain_kwargs" style syntax.
            return domain[DOMAIN_KWARGS_PARAMETER_NAME]

        return None

    if fully_qualified_parameter_name.startswith(
        DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME
    ):
        if domain and domain[DOMAIN_KWARGS_PARAMETER_NAME]:
            # Supports the "$domain.domain_kwargs.column" style syntax.
            return domain[DOMAIN_KWARGS_PARAMETER_NAME].get(
                fully_qualified_parameter_name[
                    (
                        len(
                            f"{DOMAIN_KWARGS_PARAMETER_FULLY_QUALIFIED_NAME}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}"
                        )
                    ) :
                ]
            )

        return None

    parameter_container: ParameterContainer

    if fully_qualified_parameter_name.startswith(VARIABLES_PREFIX):
        parameter_container = variables  # type: ignore[assignment] # could be None
    else:
        parameter_container = parameters[domain.id]  # type: ignore[index,union-attr] # `parameters` & `domain` could be None

    fully_qualified_parameter_name = fully_qualified_parameter_name[1:]

    fully_qualified_parameter_name_as_list: List[
        str
    ] = fully_qualified_parameter_name.split(
        FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER
    )

    if len(fully_qualified_parameter_name_as_list) == 0:
        return None

    return _get_parameter_value_from_parameter_container(
        fully_qualified_parameter_name=fully_qualified_parameter_name,
        fully_qualified_parameter_name_as_list=fully_qualified_parameter_name_as_list,
        parameter_container=parameter_container,
    )


def _get_parameter_value_from_parameter_container(
    fully_qualified_parameter_name: str,
    fully_qualified_parameter_name_as_list: List[str],
    parameter_container: ParameterContainer,
) -> Optional[Union[Any, ParameterNode]]:
    if parameter_container is None:
        return None

    parameter_node: Optional[ParameterNode] = parameter_container.get_parameter_node(
        parameter_name_root=fully_qualified_parameter_name_as_list[0]
    )
    if parameter_node is None:
        return None

    parameter_name_part: Optional[str] = None
    attribute_value_reference: Optional[str] = None
    return_value: Optional[Union[Any, ParameterNode]] = parameter_node
    parent_parameter_node: Optional[ParameterNode] = None
    try:
        for parameter_name_part in fully_qualified_parameter_name_as_list:
            parsed_attribute_name: ParseResults = _parse_attribute_naming_pattern(
                name=parameter_name_part
            )
            if len(parsed_attribute_name) < 1:
                raise KeyError(
                    f"""Unable to get value for parameter name "{fully_qualified_parameter_name}": Part \
"{parameter_name_part}" in fully-qualified parameter name does not represent a valid expression.
"""
                )

            parent_parameter_node = return_value

            attribute_value_reference = parsed_attribute_name[0]
            return_value = return_value[attribute_value_reference]  # type: ignore[index] # could be None

            parsed_attribute_name = parsed_attribute_name[1:]

            attribute_value_accessor: Union[str, int]
            for attribute_value_accessor in parsed_attribute_name:
                return_value = return_value[attribute_value_accessor]
    except KeyError:
        raise KeyError(
            f"""Unable to find value for parameter name "{fully_qualified_parameter_name}": Part \
"{parameter_name_part}" does not exist in fully-qualified parameter name.
"""
        )

    if attribute_value_reference not in parent_parameter_node:  # type: ignore[operator] # could be None
        raise KeyError(
            f"""Unable to find value for parameter name "{fully_qualified_parameter_name}": Part \
"{parameter_name_part}" of fully-qualified parameter name does not exist.
"""
        )

    return return_value


def get_parameter_values_for_fully_qualified_parameter_names(
    domain: Optional[Domain] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> Dict[str, Any]:
    fully_qualified_parameter_name: str
    return {
        fully_qualified_parameter_name: get_parameter_value_by_fully_qualified_parameter_name(
            fully_qualified_parameter_name=fully_qualified_parameter_name,
            domain=domain,
            variables=variables,
            parameters=parameters,
        )
        for fully_qualified_parameter_name in get_fully_qualified_parameter_names(
            domain=domain,
            variables=variables,
            parameters=parameters,
        )
    }


def get_fully_qualified_parameter_names(
    domain: Optional[Domain] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> List[str]:
    fully_qualified_parameter_names: List[str] = []
    if not (variables is None or variables.parameter_nodes is None):
        fully_qualified_parameter_names.extend(
            _get_parameter_node_attribute_names(
                parameter_name_root=PARAMETER_NAME_ROOT_FOR_VARIABLES,
                parameter_node=variables.parameter_nodes[
                    PARAMETER_NAME_ROOT_FOR_VARIABLES
                ],
            )
        )

    if parameters is not None:
        parameter_container: ParameterContainer = parameters[domain.id]  # type: ignore[union-attr] # could be None

        if not (
            parameter_container is None or parameter_container.parameter_nodes is None
        ):
            parameter_name_root: str
            parameter_node: ParameterNode
            for (
                parameter_name_root,
                parameter_node,
            ) in parameter_container.parameter_nodes.items():
                fully_qualified_parameter_names.extend(
                    _get_parameter_node_attribute_names(
                        parameter_name_root=PARAMETER_NAME_ROOT_FOR_PARAMETERS,
                        parameter_node=parameter_node,
                    )
                )

    return sorted(fully_qualified_parameter_names, reverse=True)


def _get_parameter_node_attribute_names(
    parameter_name_root: Optional[str] = None,
    parameter_node: Optional[ParameterNode] = None,
) -> List[str]:
    attribute_names_as_lists: List[List[str]] = []

    parameter_name_root_as_list: Optional[List[str]] = None
    if parameter_name_root:
        parameter_name_root_as_list = [parameter_name_root]

    _get_parameter_node_attribute_names_as_lists(
        attribute_names_as_lists=attribute_names_as_lists,
        parameter_name_root_as_list=parameter_name_root_as_list,
        parameter_node=parameter_node,
    )

    attribute_names: Set[str] = set()

    attribute_name: str
    for attribute_name_as_list in attribute_names_as_lists:
        attribute_name_as_list = (
            _get_parameter_name_parts_up_to_including_reserved_literal(
                attribute_name_as_list=attribute_name_as_list
            )
        )
        attribute_name = f"{FULLY_QUALIFIED_PARAMETER_NAME_DELIMITER_CHARACTER}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER.join(attribute_name_as_list[1:])}"
        attribute_names.add(attribute_name)

    return list(attribute_names)


def _get_parameter_node_attribute_names_as_lists(
    attribute_names_as_lists: List[List[str]],
    parameter_name_root_as_list: Optional[List[str]] = None,
    parameter_node: Optional[ParameterNode] = None,
) -> None:
    if parameter_node is None or parameter_name_root_as_list is None:
        return

    partial_parameter_name_root_as_list: List[str]

    attribute_name_part: str
    attribute_value_part: Any
    for attribute_name_part, attribute_value_part in parameter_node.items():
        partial_parameter_name_root_as_list = copy.deepcopy(parameter_name_root_as_list)
        partial_parameter_name_root_as_list.append(attribute_name_part)
        if isinstance(attribute_value_part, ParameterNode):
            _get_parameter_node_attribute_names_as_lists(
                attribute_names_as_lists=attribute_names_as_lists,
                parameter_name_root_as_list=partial_parameter_name_root_as_list,
                parameter_node=attribute_value_part,
            )
        else:
            attribute_names_as_lists.append(partial_parameter_name_root_as_list)


def _get_parameter_name_parts_up_to_including_reserved_literal(
    attribute_name_as_list: List[str],
) -> List[str]:
    if attribute_name_as_list[0] == PARAMETER_NAME_ROOT_FOR_VARIABLES:
        return [
            PARAMETER_NAME_ROOT_FOR_VARIABLES,
            PARAMETER_NAME_ROOT_FOR_VARIABLES,
        ]

    if not (set(attribute_name_as_list) & RESERVED_TERMINAL_LITERALS):
        return attribute_name_as_list

    # TODO: <Alex>12/29/2022: Lexicographical order avoids collisions between regular keys and reserved literals.</Alex>
    reserved_terminal_literals: List[str] = list(sorted(RESERVED_TERMINAL_LITERALS))

    idx: Optional[int] = None
    key: str
    for key in reserved_terminal_literals:
        try:
            idx = attribute_name_as_list.index(key)
            break
        except ValueError:
            pass

    return attribute_name_as_list[:idx]
